#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################
#    This package implements RC6 encryption.
#    Copyright (C) 2021, 2023, 2024  Maurice Lambert

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
###################

"""
This package implements RC6 encryption.

>>> rc6 = RC6Encryption(b'abcdefghijklmnop', 12, 32, 5)
>>> c = rc6.blocks_to_data(rc6.encrypt(b'abcdefghijklmnop'))
>>> rc6.blocks_to_data(rc6.decrypt(c))
b'abcdefghijklmnop'
>>>

>>> from urllib.request import urlopen, Request
>>> from json import dumps, load
>>> rc6 = RC6Encryption(b'abcdefghijklmnop')
>>> encrypt = rc6.blocks_to_data(rc6.encrypt(b'abcdefghijklmnop')).hex()
>>> encrypt
'bb05a4b6c46a24a8bf70e1e2c0a33e51'
>>> encrypt == load(urlopen(Request("https://www.lddgo.net/api/RC6?lang=en", headers={"Content-Type": "application/json;charset=UTF-8"}, data=dumps({"inputContent":"abcdefghijklmnop","model":"ECB","padding":"nopadding","inputPassword":"abcdefghijklmnop","inputIv":"","inputFormat":"string","outputFormat":"hex","charset":"UTF-8","encrypt":True}).encode())))["data"]
True
>>> encrypt = rc6.blocks_to_data(rc6.encrypt(b'abcdefghijklmnop') + rc6.encrypt(b'abcdefghijklmnop')).hex()
>>> encrypt
'bb05a4b6c46a24a8bf70e1e2c0a33e51bb05a4b6c46a24a8bf70e1e2c0a33e51'
>>> encrypt == load(urlopen(Request("https://www.lddgo.net/api/RC6?lang=en", headers={"Content-Type": "application/json;charset=UTF-8"}, data=dumps({"inputContent":"abcdefghijklmnopabcdefghijklmnop","model":"ECB","padding":"nopadding","inputPassword":"abcdefghijklmnop","inputIv":"","inputFormat":"string","outputFormat":"hex","charset":"UTF-8","encrypt":True}).encode())))["data"]
True
>>> pkcs5_7padding(b'abcdefghijklmno')
b'abcdefghijklmno\\x01'
>>> remove_pkcs_padding(pkcs5_7padding(b'abcdefghijklmno')) == b'abcdefghijklmno'
True
>>> pkcs5_7padding(b'abcdefghijklm')
b'abcdefghijklm\\x03\\x03\\x03'
>>> remove_pkcs_padding(pkcs5_7padding(b'abcdefghijklm')) == b'abcdefghijklm'
True
>>> rc6 = RC6Encryption(b'abcdefghijklm')
>>> encrypt = rc6.blocks_to_data(rc6.encrypt(pkcs5_7padding(b'abcdefghijklm'))).hex()
>>> encrypt
'c5ee3509788662a5711822d5e01eb4c0'
>>> encrypt == load(urlopen(Request("https://www.lddgo.net/api/RC6?lang=en", headers={"Content-Type": "application/json;charset=UTF-8"}, data=dumps({"inputContent":"abcdefghijklm","model":"ECB","padding":"pkcs5padding","inputPassword":"abcdefghijklm","inputIv":"","inputFormat":"string","outputFormat":"hex","charset":"UTF-8","encrypt":True}).encode())))["data"]
True
>>> rc6 = RC6Encryption(b'abcd')
>>> encrypt = rc6.data_encryption_ECB(b'abcdefghijklmnop').hex()
>>> encrypt
'78a64e37f7455f30aaf40750be1a065701bf2f308216c43c42c794285ffbf99e'
>>> encrypt == load(urlopen(Request("https://www.lddgo.net/api/RC6?lang=en", headers={"Content-Type": "application/json;charset=UTF-8"}, data=dumps({"inputContent":"abcdefghijklmnop","model":"ECB","padding":"pkcs5padding","inputPassword":"abcd","inputIv":"","inputFormat":"string","outputFormat":"hex","charset":"UTF-8","encrypt":True}).encode())))["data"]
True
>>> rc6.data_decryption_ECB(bytes.fromhex('78a64e37f7455f30aaf40750be1a065701bf2f308216c43c42c794285ffbf99e'))
b'abcdefghijklmnop'
>>> rc6 = RC6Encryption(b'abcd')
>>> iv, encrypt = rc6.data_encryption_CBC(b'abcdefghijklmnopabcdefghijklmnopabcdefghijklm', b'IVTEST')
>>> encrypt = encrypt.hex()
>>> iv
b'IVTESTIVTESTIVTE'
>>> encrypt
'8de91e69865825bbb6e1785e3b498f3a89708aaa2aff01a688cf9836bd7eea56c04fa0a14706d79bd94846e905bf070b'
>>> encrypt == load(urlopen(Request("https://www.lddgo.net/api/RC6?lang=en", headers={"Content-Type": "application/json;charset=UTF-8"}, data=dumps({"inputContent":"abcdefghijklmnopabcdefghijklmnopabcdefghijklm","model":"CBC","padding":"pkcs5padding","inputPassword":"abcd","inputIv":iv.decode('latin-1'),"inputFormat":"string","outputFormat":"hex","charset":"UTF-8","encrypt":True}).encode())))["data"]
True
>>> rc6.data_decryption_CBC(bytes.fromhex('8de91e69865825bbb6e1785e3b498f3a89708aaa2aff01a688cf9836bd7eea56c04fa0a14706d79bd94846e905bf070b'), iv)
b'abcdefghijklmnopabcdefghijklmnopabcdefghijklm'
>>>

~# rc6 mykey -s mydata -6
0wS4TwM292nGa378oBuz/w==
~# rc6 mykey -s 0wS4TwM292nGa378oBuz/w== -n base64 -d
mydata
~# rc6 mykey --no-sha256 -r 12 -s mydata -6
vzliI0irqi3tZ8fULxJ14g==
~# rc6 mykey --no-sha256 -r 12 -s vzliI0irqi3tZ8fULxJ14g== -n base64 -d
mydata
~# rc6 mykey -m CBC -I myiv -s mydata -1
6D7969766D7969766D7969766D796976BF2F024053F74FE27920DE5C274935A6
~# rc6 mykey -m CBC -d -s 6D7969766D7969766D7969766D796976BF2F024053F74FE27920DE5C274935A6 -n base16
mydata
~#

1 items passed all tests:
  32 tests in RC6Encryption
32 tests in 25 items.
32 passed and 0 failed.
Test passed.
"""

"""
Algorithm:

 - Key generation:
    S [0] = P32
    for i = 1 to 2r + 3 do
    {
        S [i] = S [i - 1] + Q32
    }
    A = B = i = j = 0
    v = 3 X max{c, 2r + 4}
    for s = 1 to v do
    {
        A = S [i] = (S [i] + A + B) <<< 3
        B = L [j] = (L [j] + A + B) <<< (A + B)
        i = (i + 1) mod (2r + 4)
        j = (j + 1) mod c
    }

 - Encryption:
    B = B + S[0]
    D = D + S[1]
    for i = 1 to r do
    {
        t = (B * (2B + 1)) <<< lg w
        u = (D * (2D + 1)) <<< lg w
        A = ((A ^ t) <<< u) + S[2i]
        C = ((C ^ u) <<< t) + S[2i + 1] 
        (A, B, C, D)  =  (B, C, D, A)
    }
    A = A + S[2r + 2]
    C = C + S[2r + 3]

 - Decryption:
    C = C - S[2r + 3]
    A = A - S[2r + 2]

    for i = r downto 1 do
    {
        (A, B, C, D) = (D, A, B, C)
        u = (D * (2D + 1)) <<< lg w
        t = (B * (2B + 1)) <<< lg w
        C = ((C - S[2i + 1]) >>> t) ^ u
        A = ((A - S[2i]) >>> u) ^ t
    }
    D = D - S[1]
    B = B - S[0]
"""

__version__ = "1.0.0"
__author__ = "Maurice Lambert"
__author_email__ = "mauricelambert434@gmail.com"
__maintainer__ = "Maurice Lambert"
__maintainer_email__ = "mauricelambert434@gmail.com"
__description__ = "This package implements RC6 encryption."
license = "GPL-3.0 License"
__url__ = "https://github.com/mauricelambert/RC6Encryption"

copyright = """
RC6Encryption  Copyright (C) 2021, 2023, 2024  Maurice Lambert
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.
"""
__license__ = license
__copyright__ = copyright

__all__ = ["RC6Encryption", "pkcs5_7padding"]

from base64 import (  # noqa: E402
    b85encode,
    b64encode,
    b32encode,
    b16encode,
    b85decode,
    b64decode,
    b32decode,
    b16decode,
)

from locale import getpreferredencoding  # noqa: E402
from warnings import simplefilter  # noqa: E402
from contextlib import suppress  # noqa: E402
from os import device_encoding  # noqa: E402
from functools import partial  # noqa: E402
from hashlib import sha256  # noqa: E402
from os import urandom  # noqa: E402

try:
    from binascii import a2b_hqx, b2a_hqx
except ImportError:
    uu_encoding = False
else:
    uu_encoding = True

basetwo = partial(int, base=2)
unblock = partial(int.to_bytes, length=4, byteorder="little")


class RC6Encryption:
    """
    This class implements the RC6 encryption.

    Rounds possible values: {12, 16, 20}
    """

    P32 = 0xB7E15163
    Q32 = 0x9E3779B9

    def __init__(self, key, rounds=20, w_bit=32, lgw=5):
        self.key_bytes = key
        self.rounds = rounds
        self.w_bit = w_bit
        self.lgw = lgw

        self.round2_2 = rounds * 2 + 2
        self.round2_3 = self.round2_2 + 1
        self.round2_4 = self.round2_3 + 1

        self.modulo = 2**w_bit

        (
            self.key_binary_blocks,
            self.key_integer_reverse_blocks,
        ) = self.get_blocks(key)
        self.key_blocks_number = len(self.key_binary_blocks)

        self.rc6_key = [self.P32]

        self.key_generation()

    @staticmethod
    def enumerate_blocks(data):
        """
        This function returns a tuple of 4 integers for each blocks.
        """

        _, blocks = RC6Encryption.get_blocks(data)

        while blocks:
            a, b, c, d, *blocks = blocks
            yield a, b, c, d

    @staticmethod
    def get_blocks(data):
        """
        This function returns blocks (binary strings and integers) from data.
        """

        binary_blocks = []
        integer_blocks = []
        block = ""

        for i, char in enumerate(data):
            if i and not i % 4:
                binary_blocks.append(block)
                integer_blocks.append(basetwo(block))
                block = ""

            block = "{:0>8b}".format(char) + block

        binary_blocks.append(block)
        integer_blocks.append(basetwo(block))

        return binary_blocks, integer_blocks

    @staticmethod
    def blocks_to_data(blocks):
        """
        This function returns data from blocks (binary strings).
        """

        data = b""

        for block in blocks:
            data += unblock(block)

        return data

    def right_rotation(self, x, n):
        """
        This function perform a right rotation.
        """

        mask = (2**n) - 1
        mask_bits = x & mask
        return (x >> n) | (mask_bits << (self.w_bit - n))

    def left_rotation(self, x, n):
        """
        This function perform a left rotation (based on right rotation).
        """

        return self.right_rotation(x, self.w_bit - n)

    def key_generation(self):
        """
        This function generate the key.
        """

        for i in range(0, self.round2_3):
            self.rc6_key.append((self.rc6_key[i] + self.Q32) % self.modulo)

        a = b = i = j = 0
        v = 3 * (
            self.key_blocks_number
            if self.key_blocks_number > self.round2_4
            else self.round2_4
        )

        for i_ in range(v):
            a = self.rc6_key[i] = self.left_rotation(
                (self.rc6_key[i] + a + b) % self.modulo, 3
            )
            b = self.key_integer_reverse_blocks[j] = self.left_rotation(
                (self.key_integer_reverse_blocks[j] + a + b) % self.modulo,
                (a + b) % 32,
            )
            i = (i + 1) % (self.round2_4)
            j = (j + 1) % self.key_blocks_number

        return self.rc6_key

    def data_encryption_ECB(self, data):
        """
        This function performs full encryption using ECB mode:
            - add PKCS (5/7) padding
            - get blocks
            - encrypt all blocks using ECB mode
            - convert blocks in bytes
            - returns bytes
        """

        data = pkcs5_7padding(data)
        encrypted = []

        for block in self.enumerate_blocks(data):
            encrypted.extend(self.encrypt(block))

        return self.blocks_to_data(encrypted)

    def data_decryption_ECB(self, data):
        """
        This function performs full decryption using ECB mode:
            - get blocks
            - decrypt all blocks using ECB mode
            - convert blocks in bytes
            - remove PKCS (5/7) padding
            - returns bytes
        """

        decrypted = []

        for block in self.enumerate_blocks(data):
            decrypted.extend(self.decrypt(block))

        return remove_pkcs_padding(self.blocks_to_data(decrypted))

    def data_encryption_CBC(self, data, iv=None):
        """
        This function performs full encryption using CBC mode:
            - get/generate the IV
            - add PKCS (5/7) padding
            - get blocks
            - encrypt all blocks using CBC mode
            - convert blocks in bytes
            - returns bytes
        """

        if iv is None:
            _iv = urandom(16)
        else:
            iv_length = len(iv)
            _iv = bytes(iv[i % iv_length] for i in range(16))

        _, iv = self.get_blocks(_iv)

        data = pkcs5_7padding(data)
        encrypted = []

        for block in self.enumerate_blocks(data):
            block = (
                block[0] ^ iv[0],
                block[1] ^ iv[1],
                block[2] ^ iv[2],
                block[3] ^ iv[3],
            )
            iv = self.encrypt(block)
            encrypted.extend(iv)

        return _iv, self.blocks_to_data(encrypted)

    def data_decryption_CBC(self, data, iv):
        """
        This function performs full decryption using CBC mode:
            - get blocks
            - decrypt all blocks using CBC mode
            - convert blocks in bytes
            - remove PKCS (5/7) padding
            - returns bytes
        """

        _, iv = self.get_blocks(iv)
        decrypted = []

        for block in self.enumerate_blocks(data):
            decrypted_block = self.decrypt(block)
            decrypted.extend(
                (
                    decrypted_block[0] ^ iv[0],
                    decrypted_block[1] ^ iv[1],
                    decrypted_block[2] ^ iv[2],
                    decrypted_block[3] ^ iv[3],
                )
            )
            iv = block

        return remove_pkcs_padding(self.blocks_to_data(decrypted))

    def encrypt(self, data):
        """
        This functions performs RC6 encryption on only one block.

        This function returns a list of 4 integers.
        """

        if isinstance(data, bytes):
            _, data = self.get_blocks(data)
        a, b, c, d = data

        b = (b + self.rc6_key[0]) % self.modulo
        d = (d + self.rc6_key[1]) % self.modulo

        for i in range(1, self.rounds + 1):
            t = self.left_rotation(b * (2 * b + 1) % self.modulo, self.lgw)
            u = self.left_rotation(d * (2 * d + 1) % self.modulo, self.lgw)
            tmod = t % self.w_bit
            umod = u % self.w_bit
            a = (self.left_rotation(a ^ t, umod) + self.rc6_key[2 * i]) % self.modulo
            c = (
                self.left_rotation(c ^ u, tmod) + self.rc6_key[2 * i + 1]
            ) % self.modulo
            a, b, c, d = b, c, d, a

        a = (a + self.rc6_key[self.round2_2]) % self.modulo
        c = (c + self.rc6_key[self.round2_3]) % self.modulo

        return [a, b, c, d]

    def decrypt(self, data):
        """
        This function performs a RC6 decryption.
        """

        if isinstance(data, bytes):
            _, data = self.get_blocks(data)
        a, b, c, d = data

        c = (c - self.rc6_key[self.round2_3]) % self.modulo
        a = (a - self.rc6_key[self.round2_2]) % self.modulo

        for i in range(self.rounds, 0, -1):
            (a, b, c, d) = (d, a, b, c)
            u = self.left_rotation(d * (2 * d + 1) % self.modulo, self.lgw)
            t = self.left_rotation(b * (2 * b + 1) % self.modulo, self.lgw)
            tmod = t % self.w_bit
            umod = u % self.w_bit
            c = (
                self.right_rotation((c - self.rc6_key[2 * i + 1]) % self.modulo, tmod)
                ^ u
            )
            a = self.right_rotation((a - self.rc6_key[2 * i]) % self.modulo, umod) ^ t

        d = (d - self.rc6_key[1]) % self.modulo
        b = (b - self.rc6_key[0]) % self.modulo

        return [a, b, c, d]


def remove_pkcs_padding(data):
    """
    This function implements PKCS 5/7 padding.
    """

    return data[: data[-1] * -1]


def pkcs5_7padding(data, size=16):
    """
    This function implements PKCS 5/7 padding.
    """

    pad_len = 16 - (len(data) % size)
    padding = bytes([pad_len] * pad_len)
    return data + padding


def output_encoding(data, arguments):
    """
    This function returns encoded data.
    """

    if arguments.base85 or arguments.output_encoding == "base85":
        encoding = b85encode
    elif arguments.base64 or arguments.output_encoding == "base64":
        encoding = b64encode
    elif arguments.base32 or arguments.output_encoding == "base32":
        encoding = b32encode
    elif arguments.base16 or arguments.output_encoding == "base16":
        encoding = b16encode
    elif uu_encoding and (arguments.uu or arguments.output_encoding == "uu"):
        simplefilter("ignore")
        data = b2a_hqx(data)
        simplefilter("default")
        return data
    else:
        raise ValueError("Invalid encoding algorithm value")

    return encoding(data)


def input_encoding(data, encoding):
    """
    This function returns decoded data.
    """

    if encoding == "base85":
        decoding = b85decode
    elif encoding == "base64":
        decoding = b64decode
    elif encoding == "base32":
        decoding = b32decode
    elif encoding == "base16":
        decoding = b16decode
    elif uu_encoding and encoding == "uu":
        simplefilter("ignore")
        data = a2b_hqx(data)
        simplefilter("default")
        return data
    else:
        raise ValueError("Invalid encoding algorithm value")

    return decoding(data)


def get_key(arguments):
    """
    This function returns the key (256 bits) using sha256
    by default or PKCS 5/7 for padding.
    """

    if arguments.sha256:
        return sha256(arguments.key.encode()).digest()
    else:
        return pkcs5_7padding(arguments.key.encode(), 16)[:16]


def get_data(arguments):
    """
    This function returns data for encryption from arguments.
    """

    if arguments.input_string:
        data = arguments.input_string
    else:
        data = arguments.input_file.read()

    if arguments.input_encoding:
        data = input_encoding(data, arguments.input_encoding)

    return data


def get_encodings():
    """
    This function returns the probable encodings.
    """

    encoding = getpreferredencoding()
    if encoding is not None:
        yield encoding

    encoding = device_encoding(0)
    if encoding is not None:
        yield encoding

    yield "utf-8"  # Default for Linux
    yield "cp1252"  # Default for Windows
    yield "latin-1"  # Can read all files


def decode_output(data):
    """
    This function decode outputs (try somes encoding).
    """

    output = None
    for encoding in get_encodings():
        with suppress(UnicodeDecodeError):
            output = data.decode(encoding)
            return output


"""

rc6 = RC6Encryption(b'abcd')
iv, encrypt = rc6.data_encryption_CBC(b'abcdefghijklmnopabcdefghijklmnopabcdefghijklm', b'IVTEST')



out = rc6.data_decryption_CBC(bytes.fromhex('8de91e69865825bbb6e1785e3b498f3a89708aaa2aff01a688cf9836bd7eea56c04fa0a14706d79bd94846e905bf070b'), iv)

print(out)
#b'abcdefghijklmnopabcdefghijklmnopabcdefghijklm'
"""



"""RFC 791 IPv4 header pack/parse + 1's-complement checksum.

Pure-protocol code -- no ctypes, importable everywhere.

Wire format (no options):
    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Ver=4| IHL |TOS/DSCP|         Total Length                    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |        Identification         |Flags|     Fragment Offset    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |    TTL        |   Protocol    |        Header Checksum        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                 Source Address                                |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |              Destination Address                              |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

References:
  - RFC 791 -- Internet Protocol:
    https://datatracker.ietf.org/doc/html/rfc791
  - RFC 1071 -- 1's-complement checksum algorithm:
    https://datatracker.ietf.org/doc/html/rfc1071
"""
import socket
import struct


IPV4_HDR_LEN = 20  # no options

PROTO_ICMP = 1
PROTO_TCP = 6
PROTO_UDP = 17

FLAG_DF = 0x4000  # Don't Fragment
FLAG_MF = 0x2000  # More Fragments


def checksum16(data):
    """RFC 1071 1's-complement checksum over `data` bytes.

    Returns a 16-bit integer suitable to drop into the IPv4 header
    checksum field (or the TCP/UDP segment checksum field once you've
    pre-computed it with the pseudo-header).
    """
    buf = bytes(data)
    if len(buf) & 1:
        buf = buf + b"\x00"  # zero-pad to a 16-bit boundary
    total = 0
    for i in range(0, len(buf), 2):
        total += (buf[i] << 8) | buf[i + 1]
        # Fold carry every iteration so we never overflow 32-bit on Py2;
        # cheap on Py3 too.
        total = (total & 0xffff) + (total >> 16)
    # Final fold + 1's-complement.
    while total >> 16:
        total = (total & 0xffff) + (total >> 16)
    return (~total) & 0xffff


def pack_ipv4(src_ip, dst_ip, proto, payload, ident=0, ttl=64,
              tos=0, flags=FLAG_DF, frag_offset=0):
    """Pack an IPv4 header + payload.

    src_ip / dst_ip accept dotted-quad strings or 4-byte addresses.
    payload is the L4 segment bytes (TCP/UDP/ICMP); pack_ipv4 fills in
    the Total Length, Header Checksum, and emits the full IP datagram.
    """
    if isinstance(src_ip, str):
        src = socket.inet_aton(src_ip)
    else:
        src = bytes(src_ip)
    if isinstance(dst_ip, str):
        dst = socket.inet_aton(dst_ip)
    else:
        dst = bytes(dst_ip)
    if len(src) != 4 or len(dst) != 4:
        raise ValueError("ipv4 addresses must be 4 bytes")

    total_len = IPV4_HDR_LEN + len(payload)
    ver_ihl = (4 << 4) | 5  # IHL=5 -> 20 bytes
    frag_field = (int(flags) & 0xe000) | (int(frag_offset) & 0x1fff)

    # Build with checksum=0 first, compute, patch.
    header = struct.pack(
        "!BBHHHBBH4s4s",
        ver_ihl, int(tos), total_len, int(ident) & 0xffff,
        frag_field, int(ttl), int(proto), 0, src, dst,
    )
    csum = checksum16(header)
    header = struct.pack(
        "!BBHHHBBH4s4s",
        ver_ihl, int(tos), total_len, int(ident) & 0xffff,
        frag_field, int(ttl), int(proto), csum, src, dst,
    )
    return header + bytes(payload)


class Ipv4Header(object):
    """Parsed IPv4 header.  Field names mirror RFC 791."""

    __slots__ = (
        "version", "ihl", "tos", "total_len", "ident", "flags",
        "frag_offset", "ttl", "proto", "checksum", "src", "dst",
        "header_len", "payload_offset", "raw_header",
    )

    def __init__(self, version, ihl, tos, total_len, ident, flags,
                 frag_offset, ttl, proto, checksum, src, dst,
                 header_len, payload_offset, raw_header):
        self.version = version
        self.ihl = ihl
        self.tos = tos
        self.total_len = total_len
        self.ident = ident
        self.flags = flags
        self.frag_offset = frag_offset
        self.ttl = ttl
        self.proto = proto
        self.checksum = checksum
        self.src = src
        self.dst = dst
        self.header_len = header_len
        self.payload_offset = payload_offset
        self.raw_header = raw_header

    @property
    def src_str(self):
        return socket.inet_ntoa(self.src)

    @property
    def dst_str(self):
        return socket.inet_ntoa(self.dst)


def parse_ipv4(packet_bytes):
    """Return (Ipv4Header, payload_bytes) from a complete IP datagram.

    Raises ValueError on malformed input.  Does NOT validate the
    checksum -- caller can call validate_checksum() if desired.
    """
    buf = bytes(packet_bytes)
    if len(buf) < IPV4_HDR_LEN:
        raise ValueError("ipv4 packet shorter than {0} bytes".format(IPV4_HDR_LEN))
    ver_ihl = buf[0]
    version = (ver_ihl >> 4) & 0xf
    ihl = ver_ihl & 0xf
    if version != 4:
        raise ValueError("not an IPv4 packet (version={0})".format(version))
    if ihl < 5:
        raise ValueError("invalid IHL={0}".format(ihl))
    header_len = ihl * 4
    if len(buf) < header_len:
        raise ValueError("ipv4 packet truncated mid-header")

    (_, tos, total_len, ident, frag_field,
     ttl, proto, csum, src, dst) = struct.unpack(
        "!BBHHHBBH4s4s", buf[:IPV4_HDR_LEN])
    flags = frag_field & 0xe000
    frag_offset = frag_field & 0x1fff
    raw_header = buf[:header_len]
    # Trim to total_len so we don't pass loopback FCS padding into TCP.
    if total_len < header_len:
        raise ValueError("ipv4 total_len {0} < header_len {1}".format(
            total_len, header_len))
    payload = buf[header_len:total_len] if total_len <= len(buf) else buf[header_len:]
    hdr = Ipv4Header(
        version=version, ihl=ihl, tos=tos, total_len=total_len,
        ident=ident, flags=flags, frag_offset=frag_offset, ttl=ttl,
        proto=proto, checksum=csum, src=src, dst=dst,
        header_len=header_len, payload_offset=header_len, raw_header=raw_header,
    )
    return hdr, payload


def validate_checksum(packet_bytes):
    """True iff the IPv4 header checksum verifies."""
    buf = bytes(packet_bytes)
    if len(buf) < IPV4_HDR_LEN:
        return False
    ihl = buf[0] & 0xf
    return checksum16(buf[: ihl * 4]) == 0


def pseudo_header_ipv4(src_ip, dst_ip, proto, l4_length):
    """Build the 12-byte IPv4 pseudo-header used by TCP/UDP checksums.

    See RFC 793 section 3.1 (TCP) / RFC 768 (UDP): the L4 checksum is
    computed over (pseudo-header || L4 segment).
    """
    if isinstance(src_ip, str):
        src = socket.inet_aton(src_ip)
    else:
        src = bytes(src_ip)
    if isinstance(dst_ip, str):
        dst = socket.inet_aton(dst_ip)
    else:
        dst = bytes(dst_ip)
    return struct.pack("!4s4sBBH", src, dst, 0, int(proto), int(l4_length))



"""Indexed positional string formatter and legacy eval-based formatter."""

import inspect
import re

__all__ = ["fstr", "fstr2"]


def fstr(expr, params=()):
    """Replace each {N} placeholder in expr with params[N] and return the result."""
    # Replace each {expression} with the variable value.
    def replacer(match):
        """Return str(params[index]) for the matched {index} placeholder."""
        index = int(match.group(1))
        try:
            return str(params[index])
        except (IndexError, TypeError) as e:
            raise ValueError("Error evaluating expression " + str(index) + str(e))

    out = re.sub(r"\{([^}]*)\}", replacer, expr)
    return out


class fstr2(object):
    """Legacy string formatter that uses eval() to expand {expr} placeholders.

    DEPRECATED — do not use in new code.  Use Python f-strings or fstr()
    instead.  This class is retained only for backward compatibility with
    any existing callers.

    Security note: clean_and_eval() calls eval() against the caller's local
    and global namespaces.  Never pass untrusted input to this class.

    Attributes:
        _string: the original template string.
        text: the formatted string with all {expr} placeholders expanded.
    """

    _regex = re.compile(r"\{([^{}]+)\}", re.S)

    def __init__(self, s, regex=None):
        """Init `F` with string `s`"""
        import warnings

        warnings.warn(
            "fstr2 is deprecated and uses eval() — use f-strings or fstr() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.regex = regex or self._regex
        self._string = s
        self.f_locals = self.original_caller.f_locals
        self.f_globals = self.original_caller.f_globals
        self.text = self.find_and_replace(s)

    @property
    def original_caller(self):
        names = []
        frames = []
        frame = inspect.currentframe()
        while True:
            try:
                frame = frame.f_back
                name = frame.f_code.co_name
                names.append(name)
                frames.append(frame)
            except AttributeError:
                break
        return frames[-2]

    def find_and_replace(self, s):
        """Evaluates and returns all occurrences of `regex` in `s`"""
        return re.sub(self._regex, self.clean_and_eval, s)

    def clean_and_eval(self, m):
        """Remove surrounding braces and whitespace from regex match `m`,
        evaluate, and return the result as a string.

        """
        replaced = m.group()[1:][:-1].strip()
        try:
            result = str(eval(replaced))
            return result
        except (TypeError, NameError, SyntaxError):
            try:
                result = str(eval(replaced, self.f_locals, self.f_globals))
                return result
            except (TypeError, NameError, SyntaxError):
                raise ValueError("Can't find replacement for { %s }, sorry." % replaced)

    def __str__(self):
        return str(self.text)

    def __repr__(self):
        return str(self._string)



from aionetiface import *
from aionetiface.testing import AsyncTestCase


class TestIPRange(AsyncTestCase):
    async def test_private_for_single_host_v6(self):
        ipr = IPRange("::1", bitlen=0)
        self.assertTrue(ipr.is_private)

    async def test_ip_range_iter(self):
        ipr = IPRange("192.168.0.0", bitlen=16)

        # Test iterating the range.
        i = 0
        for x in ipr:
            i += 1
            if i >= 2:
                break

    async def test_single_v4_no_params(self):
        ip = "10.0.0.1"
        ipr = IPRange(ip)
        self.assertEqual(len(ipr), 1)
        self.assertEqual(str(ipr[0]), ip)
        self.assertEqual(ipr.bitlen, 0)
        self.assertTrue(ipr.is_private)
        self.assertFalse(ipr.is_public)

    async def test_single_v4_addr_any_no_params(self):
        ip = "0.0.0.0"
        ipr = IPRange(ip)
        self.assertEqual(str(ipr[0]), ip)
        self.assertEqual(len(ipr), 1)
        self.assertTrue(ipr.is_public)
        self.assertFalse(ipr.is_private)

    async def test_single_v6_addr_any_no_params(self):
        ip = "::"
        ipr = IPRange(ip)
        self.assertEqual(str(ipr[0]), ip)
        self.assertEqual(len(ipr), 1)
        self.assertEqual(ipr.bitlen, 0)

    async def test_single_v6_link_local(self):
        ip = "fe80::ae1f:6bff:fe94:531a"
        ipr = IPRange(ip)
        self.assertEqual(str(ipr[0]), ip)
        self.assertEqual(len(ipr), 1)
        self.assertEqual(ipr.bitlen, 0)
        self.assertTrue(ipr.is_private)
        self.assertFalse(ipr.is_public)

    async def test_single_v4_broadcast(self):
        ip = "255.255.255.255"
        ipr = IPRange(ip)
        self.assertEqual(str(ipr[0]), ip)
        self.assertEqual(len(ipr), 1)
        self.assertEqual(ipr.bitlen, 0)
        self.assertTrue(ipr.is_private)
        self.assertFalse(ipr.is_public)

    async def test_block_v6_public(self):
        ip = "2402:1f00:8101:83f::"
        bitlen = 64
        ipr = IPRange(ip, netmask=None, bitlen=bitlen)
        self.assertEqual(ipr.host_no, (2**64) - 1)

        # First host in that block.
        self.assertEqual(str(ipr[0]), "2402:1f00:8101:83f::1")
        self.assertEqual(ipr.host_no, 18446744073709551615)

        # Last address in range -- all 1s set.
        self.assertEqual(str(ipr[-1]), "2402:1f00:8101:83f:ffff:ffff:ffff:ffff")

        # First address in range -- the value 1 store in 64 bits.
        self.assertEqual(str(ipr[0]), "2402:1f00:8101:83f::1")
        self.assertEqual(str(ipr[1]), "2402:1f00:8101:83f::2")

        # Test > sys.maxsize = memory addr overflow.
        # Len is expected to overflow.
        ip = "2402:1f00:8101::"
        bitlen = 80
        ipr = IPRange(ip, netmask=None, bitlen=bitlen)
        self.assertEqual(ipr.host_no, (2**80) - 1)

        # include 0 in range = -1, start counting from 0 = -1 so -2
        self.assertEqual(
            str(ipr[(2**80) - 2]), "2402:1f00:8101:ffff:ffff:ffff:ffff:ffff"
        )
        self.assertTrue(ipr.is_public)
        self.assertFalse(ipr.is_private)

    async def test_block_v4_public(self):
        ipr = IPRange("8.8.8.0", bitlen=8)
        self.assertEqual(str(ipr[0]), "8.8.8.1")
        ipr = IPRange("8.8.8.4", bitlen=8)
        self.assertEqual(str(ipr[-1]), "8.8.8.255")
        self.assertEqual(str(ipr[256]), "8.8.8.1")

    async def test_loopback_ips(self):
        v4_lb = "127.0.0.1"
        v6_lb = "::1"
        ipr = IPRange(v4_lb)
        self.assertEqual(len(ipr), 1)
        self.assertEqual(str(ipr[0]), v4_lb)
        self.assertEqual(str(ipr[4]), v4_lb)
        ipr = IPRange(v6_lb)
        self.assertEqual(len(ipr), 1)
        self.assertEqual(str(ipr[0]), v6_lb)
        self.assertEqual(str(ipr[3]), v6_lb)

    async def test_black_hole_ips(self):
        for ip in BLACK_HOLE_IPS.values():
            ipr = IPRange(ip)
            self.assertTrue(ipr.is_public)
            self.assertFalse(ipr.is_private)
            self.assertEqual(ip_norm(str(ipr[0])), ip)

    async def test_misc_ipr(self):
        ipr = IPRange("192.168.0.0", netmask="255.255.0.0")
        n = 0
        for needle in reversed(ipr):
            n += 1
            if n >= 2:
                break

        # Test host add works.
        ipr_a = IPRange("192.168.0.0", netmask="255.255.0.0")
        ipr_b = IPRange("192.169.0.4", netmask="255.255.0.0")
        ipr_c = IPRange("192.168.0.4", netmask="255.255.0.0")
        self.assertEqual(ipr_a + ipr_b, ipr_c)
        self.assertEqual(ipr_a + 4, ipr_c)

        # Test host sub works.
        ipr_d = IPRange("192.168.255.252", netmask="255.255.0.0")
        self.assertEqual(ipr_a - ipr_b, ipr_d)
        self.assertEqual(ipr_a - 4, ipr_d)

        # Get items.
        ipr_z = IPRange("192.168.0.3")
        ipr_y = IPRange("192.168.0.4")
        ipr_x = ipr_a[3]
        self.assertEqual(ipr_x, ipr_y)
        self.assertTrue(ipr_y > ipr_z)
        self.assertTrue(ipr_z < ipr_y)

        # You can use tuples to fetch a list of sub objects.
        ipr_list = ipr_a[(0, 1, 2)]
        hey_list = [
            IPRange("192.168.0.1"),
            IPRange("192.168.0.2"),
            IPRange("192.168.0.3"),
        ]
        self.assertEqual(ipr_list, hey_list)


if __name__ == "__main__":
    main()



"""Port allocation helpers for pytest-xdist parallel test runs."""
import os


def xdist_port_base(base, stride=200):
    """Return base offset for test ports, unique per xdist worker."""
    w = os.environ.get("PYTEST_XDIST_WORKER", "")
    try:
        n = int(w.replace("gw", "")) if w else 0
    except ValueError:
        n = 0
    return base + n * stride



"""Shared libpcap ctypes binding.

Every Unix shim (linux.py, bsd.py, darwin.py) and the Windows shim
(windows.py) call into this module after locating the right shared
library / DLL.  The libpcap C ABI has been stable since libpcap 0.9
(2005) -- WinPcap 4.1.3 is libpcap 1.0-era, Npcap is libpcap 1.10-era,
modern Linux ships libpcap 1.x -- and the symbol surface we need has
not changed:

    pcap_findalldevs / pcap_freealldevs
    pcap_open_live
    pcap_close
    pcap_sendpacket
    pcap_next_ex
    pcap_compile / pcap_setfilter / pcap_freecode
    pcap_datalink
    pcap_lib_version
    pcap_geterr

That is all the userspace TCP layer needs.  Everything else (filter
optimisation, statistics, asynchronous capture loops) we either don't
care about or rebuild in pure Python.

References:
  - pcap(3PCAP)  https://www.tcpdump.org/manpages/pcap.3pcap.html
  - pcap_next_ex(3PCAP)  https://www.tcpdump.org/manpages/pcap_next_ex.3pcap.html
  - pcap_sendpacket(3PCAP)  https://www.tcpdump.org/manpages/pcap_sendpacket.3pcap.html
  - pcap_findalldevs(3PCAP)  https://www.tcpdump.org/manpages/pcap_findalldevs.3pcap.html
"""
import ctypes
from ctypes import (
    POINTER, Structure, c_char, c_char_p, c_int, c_uint, c_uint32,
    c_void_p, c_short, c_ushort, c_ubyte, byref, cast, create_string_buffer,
)
import socket as stdlib_socket

from ..backend import Backend, BackendFactory, PcapError, PcapUnavailableError


PCAP_ERRBUF_SIZE = 256

# Common DLT (data link type) values.  See:
# https://www.tcpdump.org/linktypes.html
DLT_NULL = 0
DLT_EN10MB = 1
DLT_RAW = 12   # on most platforms (BSD/Linux); see linktypes.html
DLT_LOOP = 108  # OpenBSD loopback

# pcap_next_ex return codes
PCAP_NEXT_OK = 1
PCAP_NEXT_TIMEOUT = 0
PCAP_NEXT_EOF = -2
PCAP_NEXT_ERROR = -1


class TimevalT(Structure):
    """Mirror of struct timeval as libpcap consumes it.  Field widths
    are platform-specific (32-bit vs 64-bit time_t) but for our use
    -- we only read it for ordering, never write it -- c_long is fine
    on every libpcap target we have.  The struct is opaque to us if
    we don't read those fields, so this is a defensive shape only."""
    _fields_ = (
        ("tv_sec", ctypes.c_long),
        ("tv_usec", ctypes.c_long),
    )


class PcapPktHdr(Structure):
    """Mirror of struct pcap_pkthdr -- the header libpcap prepends to
    each captured frame.  caplen is what's actually in the buffer
    (snaplen-bounded), len is the original on-wire length."""
    _fields_ = (
        ("ts", TimevalT),
        ("caplen", c_uint32),
        ("len", c_uint32),
    )


class BpfProgram(Structure):
    """struct bpf_program -- output of pcap_compile, input to
    pcap_setfilter.  We treat the contents as opaque and only pass
    pointers around."""
    _fields_ = (
        ("bf_len", c_uint),
        ("bf_insns", c_void_p),
    )


class SockaddrIn(Structure):
    """Just enough of struct sockaddr_in to pull the IPv4 address out
    of a pcap_addr.addr pointer when listing interfaces."""
    _fields_ = (
        ("sin_family", c_ushort),
        ("sin_port", c_ushort),
        ("sin_addr", c_uint32),
        ("sin_zero", c_char * 8),
    )


class SockaddrIn6(Structure):
    """Just enough of struct sockaddr_in6 for IPv6 address extraction
    out of pcap_addr.addr."""
    _fields_ = (
        ("sin6_family", c_ushort),
        ("sin6_port", c_ushort),
        ("sin6_flowinfo", c_uint32),
        ("sin6_addr", c_ubyte * 16),
        ("sin6_scope_id", c_uint32),
    )


class PcapAddr(Structure):
    pass


# Self-referential next pointer -- declared after the class body.
PcapAddr._fields_ = (
    ("next", POINTER(PcapAddr)),
    ("addr", c_void_p),
    ("netmask", c_void_p),
    ("broadaddr", c_void_p),
    ("dstaddr", c_void_p),
)


class PcapIf(Structure):
    pass


PcapIf._fields_ = (
    ("next", POINTER(PcapIf)),
    ("name", c_char_p),
    ("description", c_char_p),
    ("addresses", POINTER(PcapAddr)),
    ("flags", c_uint32),
)


PCAP_IF_LOOPBACK = 0x00000001
PCAP_IF_UP = 0x00000002


def bind_symbols(lib):
    """Attach argtypes / restype to every pcap_* symbol we touch.

    Done once per shim, after the shim has located its shared library.
    Returns the same lib so callers can chain or store it directly.
    """
    lib.pcap_lib_version.argtypes = ()
    lib.pcap_lib_version.restype = c_char_p

    lib.pcap_findalldevs.argtypes = (POINTER(POINTER(PcapIf)), c_char_p)
    lib.pcap_findalldevs.restype = c_int

    lib.pcap_freealldevs.argtypes = (POINTER(PcapIf),)
    lib.pcap_freealldevs.restype = None

    lib.pcap_open_live.argtypes = (c_char_p, c_int, c_int, c_int, c_char_p)
    lib.pcap_open_live.restype = c_void_p

    lib.pcap_close.argtypes = (c_void_p,)
    lib.pcap_close.restype = None

    lib.pcap_sendpacket.argtypes = (c_void_p, c_char_p, c_int)
    lib.pcap_sendpacket.restype = c_int

    lib.pcap_next_ex.argtypes = (
        c_void_p, POINTER(POINTER(PcapPktHdr)), POINTER(POINTER(c_ubyte))
    )
    lib.pcap_next_ex.restype = c_int

    lib.pcap_compile.argtypes = (
        c_void_p, POINTER(BpfProgram), c_char_p, c_int, c_uint32
    )
    lib.pcap_compile.restype = c_int

    lib.pcap_setfilter.argtypes = (c_void_p, POINTER(BpfProgram))
    lib.pcap_setfilter.restype = c_int

    lib.pcap_freecode.argtypes = (POINTER(BpfProgram),)
    lib.pcap_freecode.restype = None

    lib.pcap_datalink.argtypes = (c_void_p,)
    lib.pcap_datalink.restype = c_int

    lib.pcap_geterr.argtypes = (c_void_p,)
    lib.pcap_geterr.restype = c_char_p

    return lib


def decode_sockaddr(addr_ptr):
    """Read a sockaddr_in / sockaddr_in6 out of a void* and return the
    string-form address, or None on unknown family / null pointer."""
    if not addr_ptr:
        return None
    # The leading sa_family field is at offset 0 for both BSD-derived
    # and Linux sockaddr layouts (Linux uses a 16-bit sa_family; BSD
    # uses 8-bit sa_len + 8-bit sa_family which lands the family byte
    # at offset 1).  Trying both gives us cross-platform interface
    # listing without per-OS branches here.
    family_word = ctypes.cast(addr_ptr, POINTER(c_ushort))[0]
    family_low = family_word & 0xff
    family_high = (family_word >> 8) & 0xff
    candidates = (family_low, family_high, family_word)
    for fam in candidates:
        if fam == stdlib_socket.AF_INET:
            sa = ctypes.cast(addr_ptr, POINTER(SockaddrIn))[0]
            # sin_addr.s_addr is stored in network byte order in the
            # kernel struct.  ctypes' c_uint32 reads it as a host-
            # endian integer, so pack with native ("=") byte order to
            # recover the original in-memory layout.
            import struct
            packed = struct.pack("=I", int(sa.sin_addr) & 0xffffffff)
            try:
                return stdlib_socket.inet_ntop(stdlib_socket.AF_INET, packed)
            except (OSError, ValueError):
                return None
        if fam == stdlib_socket.AF_INET6:
            sa = ctypes.cast(addr_ptr, POINTER(SockaddrIn6))[0]
            packed = bytes(bytearray(sa.sin6_addr))
            try:
                return stdlib_socket.inet_ntop(stdlib_socket.AF_INET6, packed)
            except (OSError, ValueError):
                return None
    return None


class LibpcapBackend(Backend):
    """Backend instance wrapping one pcap_t handle.

    Concrete shims share this class and only differ in how they
    located libpcap.  `lib` is the bound ctypes.CDLL; `handle` is the
    pcap_t pointer returned by pcap_open_live.
    """

    def __init__(self, lib, handle, iface_name, datalink_value):
        self.lib = lib
        self.handle = handle
        self.iface_name = iface_name
        self.datalink_value = datalink_value
        self.closed = False
        self.platform_name = "libpcap"

    def datalink(self):
        return self.datalink_value

    def send(self, frame_bytes):
        if self.closed:
            raise PcapError("send on closed pcap handle")
        if not isinstance(frame_bytes, (bytes, bytearray)):
            raise ValueError("frame_bytes must be bytes or bytearray")
        buf = bytes(frame_bytes)
        rc = self.lib.pcap_sendpacket(self.handle, buf, len(buf))
        if rc != 0:
            err = self.lib.pcap_geterr(self.handle)
            err_text = err.decode("ascii", "replace") if err else "unknown"
            raise PcapError("pcap_sendpacket failed: {0}".format(err_text))
        return len(buf)

    def recv(self, timeout_ms=100):
        """Wraps pcap_next_ex.  Returns bytes or None on timeout.

        timeout_ms is largely advisory -- libpcap's per-call timeout is
        the one set at pcap_open_live time.  We loop briefly inside so
        a caller-provided shorter window still feels responsive even
        when the underlying read-timeout is 10 ms (our default for
        open_live)."""
        if self.closed:
            raise PcapError("recv on closed pcap handle")
        hdr_ptr = POINTER(PcapPktHdr)()
        data_ptr = POINTER(c_ubyte)()
        rc = self.lib.pcap_next_ex(self.handle, byref(hdr_ptr), byref(data_ptr))
        if rc == PCAP_NEXT_OK:
            caplen = hdr_ptr[0].caplen
            if caplen == 0:
                return b""
            return ctypes.string_at(data_ptr, caplen)
        if rc == PCAP_NEXT_TIMEOUT:
            return None
        if rc == PCAP_NEXT_EOF:
            raise PcapError("pcap_next_ex: end of capture (savefile EOF)")
        # rc == -1 or unknown
        err = self.lib.pcap_geterr(self.handle)
        err_text = err.decode("ascii", "replace") if err else "unknown"
        raise PcapError("pcap_next_ex failed: {0}".format(err_text))

    def set_filter(self, bpf_string):
        if self.closed:
            raise PcapError("set_filter on closed pcap handle")
        prog = BpfProgram()
        text = bpf_string.encode("ascii") if bpf_string else b""
        # netmask=0 is acceptable when we are not filtering by network
        # mask (pcap docs say PCAP_NETMASK_UNKNOWN is fine for live).
        rc = self.lib.pcap_compile(self.handle, byref(prog), text, 1, 0)
        if rc != 0:
            err = self.lib.pcap_geterr(self.handle)
            err_text = err.decode("ascii", "replace") if err else "unknown"
            raise PcapError("pcap_compile({0!r}) failed: {1}".format(
                bpf_string, err_text))
        try:
            rc = self.lib.pcap_setfilter(self.handle, byref(prog))
            if rc != 0:
                err = self.lib.pcap_geterr(self.handle)
                err_text = err.decode("ascii", "replace") if err else "unknown"
                raise PcapError("pcap_setfilter failed: {0}".format(err_text))
        finally:
            self.lib.pcap_freecode(byref(prog))

    def close(self):
        if self.closed:
            return
        try:
            self.lib.pcap_close(self.handle)
        finally:
            self.closed = True
            self.handle = None


class LibpcapFactory(BackendFactory):
    """Shared factory base used by every Unix shim and by Windows
    once it's loaded wpcap.dll.  Concrete subclasses override the
    `load_lib()` hook to return either the bound CDLL or None."""

    platform_name = "libpcap"
    library_path_label = "libpcap"

    def __init__(self):
        self.lib = None
        self.load_attempted = False

    def load_lib(self):
        """Return a bound CDLL or None.  Subclasses do platform-specific
        discovery here -- we never try to second-guess where libpcap
        lives on disk from inside this base class."""
        return None

    def ensure_loaded(self):
        if self.load_attempted:
            return self.lib is not None
        self.load_attempted = True
        try:
            self.lib = self.load_lib()
        except OSError as exc:
            self.lib = None
        return self.lib is not None

    def available(self):
        return self.ensure_loaded()

    def library_version(self):
        if not self.ensure_loaded():
            return None
        ver = self.lib.pcap_lib_version()
        if ver is None:
            return None
        return ver.decode("ascii", "replace")

    def list_interfaces(self):
        if not self.ensure_loaded():
            raise PcapUnavailableError("pcap library not loaded")
        errbuf = create_string_buffer(PCAP_ERRBUF_SIZE)
        head_ptr = POINTER(PcapIf)()
        rc = self.lib.pcap_findalldevs(byref(head_ptr), errbuf)
        if rc != 0:
            raise PcapError("pcap_findalldevs failed: {0}".format(
                errbuf.value.decode("ascii", "replace")))
        out = []
        try:
            cur = head_ptr
            while cur:
                node = cur[0]
                addrs = []
                acur = node.addresses
                while acur:
                    a = acur[0]
                    val = decode_sockaddr(a.addr)
                    if val:
                        addrs.append(val)
                    acur = a.next
                name = node.name.decode("ascii", "replace") if node.name else ""
                desc = (node.description.decode("ascii", "replace")
                        if node.description else "")
                out.append({
                    "name": name,
                    "description": desc,
                    "loopback": bool(node.flags & PCAP_IF_LOOPBACK),
                    "up": bool(node.flags & PCAP_IF_UP),
                    "addresses": addrs,
                })
                cur = node.next
        finally:
            self.lib.pcap_freealldevs(head_ptr)
        return out

    def open(self, iface_name, snaplen=65535, promisc=False, timeout_ms=10):
        if not self.ensure_loaded():
            raise PcapUnavailableError("pcap library not loaded")
        errbuf = create_string_buffer(PCAP_ERRBUF_SIZE)
        name_bytes = iface_name.encode("utf-8") if isinstance(iface_name, str) else iface_name
        handle = self.lib.pcap_open_live(
            name_bytes,
            int(snaplen),
            1 if promisc else 0,
            int(timeout_ms),
            errbuf,
        )
        if not handle:
            raise PcapError("pcap_open_live({0!r}) failed: {1}".format(
                iface_name, errbuf.value.decode("ascii", "replace")))
        dlt = self.lib.pcap_datalink(handle)
        return LibpcapBackend(self.lib, handle, iface_name, dlt)



"""Platform-specific pcap shims.

Each module in this package binds the libpcap ABI for one OS family
via ctypes and exports a module-level `factory` (BackendFactory
instance).  The top-level `backend.get_backend()` picks the right one
based on sys.platform.

Why ctypes-direct and not pcapy / pylibpcap / scapy:
  - pcapy is a C extension and ships no wheel for Py3.5 / Windows XP.
  - pylibpcap is similarly out-of-date and abandoned upstream.
  - scapy is far heavier than the four-or-five libpcap symbols we
    actually need and pulls in optional crypto deps.
  - The libpcap C ABI has been stable since libpcap 0.9 (~2005),
    which covers every system in our test matrix.

References:
  - libpcap manpages live at https://www.tcpdump.org/manpages/
  - WinPcap developer pack docs:
    https://www.winpcap.org/docs/docs_412/html/group__wpcap.html
  - Npcap is API-compatible with WinPcap by design:
    https://npcap.com/guide/wpcap/intro.html
"""



"""
This module is a drop-in replacement for netifaces on Windows.
Usage is simply:
    from aionetiface import *
    async def main():
        netifaces = await aionetiface_setup_netifaces()

The pypi netifaces module has several problems on Windows OS':

1. It requires a distribution of VS C++ Build Tools 20**.
Having users install the right build tools to get the software
working is complex and error-prone. It also makes packaging
software that uses netifaces much more difficult.

2. On recent versions of Windows it displays meaningless
strings of GUIDs over network interface names. Consequently,
one has to do a series of reg hacks, privilege elevations,
and patches to use the library with human-readable names.

The most portable way to do complex operations on Windows
seems to be to use powershell scripting. Looking up certain
registry keys isn't guaranteed to work as the locations may
change between versions. The networking tools available in
cmd.exe are now being phased out in favor of the tools
available in powershell. Powershell is now widely supported
even on older Windows OS.

The downside to using powershell to obtain relevant NIC
information is it's slow. A process needs to be spawned
for each new command. In order to prevent errors on certain
Windows versions concurrency also has to be disabled. This
doesn't seem to effect speed too much but the code is many
times slower than netiface. Nevertheless, it's more
portable and doesn't require privilege elevations to get
human-readable network interface descriptions.

Speedups:

I've added a new feature to obtain all the networking info
from a single powershell script. The program will attempt
to use this script if powershell is unrestricted. If it
fails to load interfaces with regular commands it will
try relaunch with a UAC prompt, unrestrict powershell,
then attempt to run the script in powershell.

Registry:

Don't attempt to write a new version of this using the
Windows registry. The registry does not make available
simple access to the interface numbers and a special
(undocumented) algorithm needs to be used to rank the
interfaces and derive the number. Doing this is unsupported
outside of using heavy Windows SDK dependencies.

Notes:
    - These commands don't seem to require special permissions.
    - They need to use double quotes or the command won't run.
    - Tested as working with execution policy = restricted.
    - Is there a way to convert the PS1 script to a single line
    command and have it passed to powershell without execution perms?
"""

import asyncio
import re
import platform
from ....net.net_defs import IP4, IP6, VALID_AFS


def is_loopback_iface(if_info):
    """True iff every IP on this adapter is a loopback address.

    Windows surfaces a "Loopback Pseudo-Interface 1" via netsh / WMIC /
    iphlpapi the same way it surfaces a real NIC.  Its only IPs are
    127.0.0.1 and ::1.  An interface is treated as loopback iff it has
    at least one address and every address is in 127.0.0.0/8 or equals
    ::1 -- a real adapter that happens to alias 127.x is impossible on
    Windows, so the check is unambiguous.

    Returns False for an interface with zero addresses (disabled NICs
    still appear in the index; those get filtered at a different layer
    so don't shadow them here).
    """
    addrs = if_info.get("addr") or {}
    saw_any = False
    for af in (IP4, IP6):
        for entry in addrs.get(af, []):
            ip = entry.get("addr") if isinstance(entry, dict) else None
            if not ip:
                continue
            saw_any = True
            ip_only = ip.split("/")[0]
            if af == IP4 and not ip_only.startswith("127."):
                return False
            if af == IP6 and ip_only != "::1":
                return False
    return saw_any
from ....net.net_utils import cidr_to_netmask, v_to_af, ip_strip_if, ip_strip_cidr, af_bitlen
from ....net.ip_range import IPRange
from ....utility.utils import fstr, log, log_exception, to_n, ip_f
from ....utility.cmd_tools import cmd, get_powershell_path, ps1_exec_trick
from .win_netsh import if_infos_from_netsh
from .win_iphlpapi import if_infos_from_iphlpapi, is_supported as iphlpapi_is_supported
from .win_wmic import if_infos_from_wmic


CMD_TIMEOUT = 30

IFS_PS1 = """
# Load default interface for IPv4 and IPv6.
$v4_default = Find-NetRoute -RemoteIPAddress 0.0.0.0 -erroraction 'silentlycontinue' | Format-List -Property ifIndex
$v6_default = Find-NetRoute -RemoteIPAddress :: -erroraction 'silentlycontinue' | Format-List -Property ifIndex
if($v4_default -eq $null){
    $v4_default = "null"
}
if($v6_default -eq $null){
    $v6_default = "null"
}

# Show them if any.
echo("4444444444")
echo($v4_default)
echo("4444444444")
echo("6666666666")
echo($v6_default)
echo("6666666666")

# Load interfaces and associated addresses. removed -physical
$ifs = Get-NetAdapter -erroraction 'silentlycontinue' | where status -eq up 
Foreach($iface in $ifs){
    # Get first hop for the iface for both AFs.
    $v4gw = (Get-NetRoute "0.0.0.0/0" -InterfaceIndex $iface.ifIndex  -erroraction 'silentlycontinue').NextHop 
    $v6gw = (Get-NetRoute "::/0" -InterfaceIndex $iface.ifIndex  -erroraction 'silentlycontinue').NextHop
    if($v4gw -eq $null){
        $v4gw = "null"
    }

    if($v6gw -eq $null){
        $v6gw = "null"
    }

    # Build a list of IPs with prefix lengths (e.g. "192.168.1.5/24").
    $ips = @()
    $addrs = Get-NetIPAddress -InterfaceIndex $iface.ifIndex
    Foreach($addr in $addrs){
        $ips += $addr.IPAddress + "/" + $addr.PrefixLength
    }

    # Save them as new property values.
    $iface | Add-Member -NotePropertyName v4GW -NotePropertyValue $v4gw
    $iface | Add-Member -NotePropertyName v6GW -NotePropertyValue $v6gw

    # Output this interface info with it's addresses.
    $out = $iface | Format-List -Property InterfaceDescription,ifIndex,InterfaceGuid,MacAddress,v4GW,v6GW
    echo($out)
    echo($ips)
}
"""


async def load_ifs_from_ps1():
    # Get all interface details as one big script.
    out = await asyncio.wait_for(ps1_exec_trick(IFS_PS1), 10)
    if "InterfaceDescription" not in out:
        raise Exception("Invalid powershell output for net script.")
    if "MethodNotFound" in out:
        raise Exception("Unknown error with powershell.")
    if "is not recognized as the name" in out:
        raise Exception("Powershell doesn't support these features.")

    # Load default interface by if_index.
    default_ifs = {IP4: None, IP6: None}
    if_defaults_by_index = {}
    for v in [4, 6]:
        # No to AF.
        af = v_to_af(v)

        # Regex to extract the interface no for the AF.
        delim = str(v) * 10
        p = fstr(
            r"{0}[\s\S]*ifIndex\s*:\s*([0-9]+)[\s\S]*{1}",
            (
                delim,
                delim,
            ),
        )
        if_index = re.findall(p, out)

        # Save it in a loopup table.
        if len(if_index):
            if_index = to_n(if_index[0])

            # Save it by AF.
            default_ifs[af] = if_index

            # Make a list to store AF by if_index.
            if if_index not in if_defaults_by_index:
                if_defaults_by_index[if_index] = []

            # Save AF by if_index.
            if_defaults_by_index[if_index].append(af)

    # Extract interface details.
    p = r"InterfaceDescription *: *([^\r\n]+?) *[\r\n]+ifIndex *: *([0-9]+?) *[\r\n]+InterfaceGuid *: *([^\r\n]+?) *[\r\n]+MacAddress *: *([^\r\n]+?) *[\r\n]+v4GW *: *([^\r\n]+?) *[\r\n]+v6GW *: *([^ ]+?)[\s]+((?:[0-9a-f.:%/]+ *[\r\n]*)+)"
    re_results = re.findall(p, out)

    # Index the results into a dict.
    if_infos = []
    if len(re_results):
        for r in re_results:
            if_index = to_n(r[1])
            if_info = {
                "guid": r[2],
                "name": r[0],
                "no": if_index,
                "mac": r[3],
                # Placeholders.
                "addr": None,
                "gws": {
                    IP4: None if r[4] == "null" else r[4],
                    IP6: None if r[5] == "null" else r[5],
                },
                "defaults": [],
            }

            # Set defaults.
            if if_index in if_defaults_by_index:
                if_info["defaults"] = if_defaults_by_index[if_index]

            # Process address info.
            addr_info = {IP4: [], IP6: []}
            addr_s = r[6].replace(" ", "")
            addr_list = addr_s.splitlines(False)
            for addr in addr_list:
                # Skip blank entries.
                if addr == "":
                    continue

                # Extract prefix length from "ip/prefix" format added by the PS1 script.
                prefix_len = None
                if "/" in addr:
                    raw_ip, raw_prefix = addr.rsplit("/", 1)
                    try:
                        prefix_len = int(raw_prefix)
                    except ValueError:
                        pass
                    addr = raw_ip

                addr = ip_strip_cidr(ip_strip_if(addr))
                ipr = IPRange(addr)
                host_limit = prefix_len if prefix_len is not None else af_bitlen(ipr.af)
                addr_info[ipr.af].append(
                    {
                        "addr": addr,
                        "af": ipr.af,
                        "host_limit": host_limit,
                        "netmask": cidr_to_netmask(host_limit, ipr.af),
                    }
                )

            # Save addresses.
            if_info["addr"] = addr_info

            # Save info.
            if_infos.append(if_info)

    """
    If no route found for a given address family set the first
    interface as the default interface for that AF.
    """
    for af in VALID_AFS:
        if default_ifs[af] is None:
            if len(if_infos):
                if if_infos[0]["gws"][af] is not None:
                    if_infos[0]["defaults"].append(af)

    return if_infos


async def get_default_gw_by_if_index(af, if_index):
    dest_ip = "0.0.0.0/0" if af == IP4 else "::/0"
    cmd_str = '{} "(Get-NetRoute {} -InterfaceIndex {}).NextHop"'
    cmd_str = cmd_str.format("powershell", dest_ip, if_index)

    # Execute the command.
    try:
        out = await cmd(cmd_str, timeout=CMD_TIMEOUT)
    except (OSError, asyncio.TimeoutError):
        return None

    if out is None:
        return None

    # Return the string if it's a valid IP.
    out = out.strip()
    try:
        ip_f(out)
        return out
    except ValueError:
        log(out)
        return None


async def get_addr_info_by_if_index(if_index):
    addr = {IP4: [], IP6: []}
    cmd_str = 'powershell "Get-NetIPAddress -InterfaceIndex {}"'
    cmd_str = cmd_str.format(if_index)
    try:
        out = await cmd(cmd_str, timeout=CMD_TIMEOUT)
    except (OSError, asyncio.TimeoutError):
        return addr

    try:
        addr_infos = re.findall(
            r"IPAddress\s*:\s*([^\s]*)[\s\S]*?AddressFamily\s*:\s*([^\s]+)[\s\S]*?PrefixLength\s*:\s([0-9]+)",
            out,
        )

        for addr_info in addr_infos:
            ip_val, af_family, host_limit = addr_info
            host_limit = int(host_limit)
            if af_family == "IPv4":
                af = IP4
            if af_family == "IPv6":
                af = IP6

            addr[af].append(
                {
                    "addr": ip_val,
                    "af": af,
                    "host_limit": host_limit,
                    "netmask": cidr_to_netmask(host_limit, af),
                }
            )
    except (ValueError, re.error):
        log_exception()
        return addr

    return addr


async def get_default_iface_by_af(af):
    if af == IP4:
        any_offset = 0
    if af == IP6:
        any_offset = 1

    any_addr_list = ["0.0.0.0", "::"]
    dest_ip = any_addr_list[any_offset]
    cmd_buf = 'powershell "Find-NetRoute -RemoteIPAddress {}"'
    cmd_buf = cmd_buf.format(dest_ip)
    try:
        out = await cmd(cmd_buf, timeout=CMD_TIMEOUT)
    except (OSError, asyncio.TimeoutError):
        return None

    try:
        if_index_str = re.findall(r"InterfaceIndex\s*:\s*([0-9]+)", out)
        if len(if_index_str):
            return int(if_index_str[0])
        else:
            # If an AF is not support an error is thrown
            # and the pattern above won't match anything.
            return None
    except (ValueError, re.error):
        log_exception()
        return None


def extract_if_fields(ifs_str):
    """Returns a list of interface field dicts parsed from powershell Get-NetAdapter output."""
    results = []
    try:
        if_info_matches = re.findall(
            r"InterfaceDescription\s*:\s([^\r\n]*?)[\r\n]+ifIndex\s*:\s*([0-9]+)\s*InterfaceGuid\s*:\s*([^\r\n]+)\s*MacAddress\s*:\s*([^\s]+)\s*",
            ifs_str,
        )  # noqa
        if len(if_info_matches):
            for if_info_match in if_info_matches:
                if_desc, if_index, guid, mac_addr = if_info_match
                if_index = int(if_index)
                results.append(
                    {
                        "guid": guid,
                        "name": if_desc,
                        "no": if_index,
                        "mac": mac_addr,
                        # Placeholders.
                        "addr": None,
                        "gws": {IP4: None, IP6: None},
                        "defaults": None,
                    }
                )
    except (ValueError, re.error, AttributeError):
        log_exception()
        return results

    return results


# Get list of net adaptors via powershell.
# Ignore hidden adapters. Non-physical or down.
# Specify desc and index to show full entry.
async def get_ifaces():
    ps_path = get_powershell_path()
    try:
        out = await asyncio.wait_for(
            cmd(
                fstr(
                    '{0} "Get-NetAdapter -physical | where status -eq up  | Format-List -Property InterfaceDescription,ifIndex,InterfaceGuid,MacAddress"',
                    (ps_path,),
                )
            ),
            CMD_TIMEOUT,
        )
        if "InterfaceDescription" not in out:
            raise Exception("Get net adapter error.")
    except (OSError, asyncio.TimeoutError):
        log_exception()
        out = ""

    return out


async def win_load_interface_state(if_results):
    # Lookup whether an1
    if_defaults_by_index = {}
    af_index = {}

    async def set_ip4_if_index():
        af_index[IP4] = await get_default_iface_by_af(IP4)

    async def set_ip6_if_index():
        af_index[IP6] = await get_default_iface_by_af(IP6)

    # Execute the above functions.
    tasks = [
        set_ip4_if_index(),
        set_ip6_if_index(),
    ]
    for task in tasks:
        await task

    # Set the AFs the interface is the default for.
    if af_index[IP4] or af_index[IP6]:
        if af_index[IP4]:
            if_defaults_by_index[af_index[IP4]] = [IP4]

        if af_index[IP6]:
            if_defaults_by_index[af_index[IP6]] = [IP6]

        if af_index[IP4] == af_index[IP6]:
            if_defaults_by_index[af_index[IP4]] = [IP4, IP6]

    # Parse output lines from Get-Adapter.
    if_tasks = []
    by_guid_index = {}
    for result in if_results:

        async def if_task_func():
            try:
                if_index = result["no"]
                guid = result["guid"]

                # Default interface for these address families.
                default_for_afs = []
                if if_index in if_defaults_by_index:
                    default_for_afs = if_defaults_by_index[if_index]

                result["defaults"] = default_for_afs

                # Otherwise it's worth saving.
                by_guid_index[guid] = result

                # Get address information.
                async def set_addr():
                    by_guid_index[guid]["addr"] = await get_addr_info_by_if_index(
                        if_index
                    )

                # Get default gateways.
                async def set_ip4_gw():
                    by_guid_index[guid]["gws"][IP4] = await get_default_gw_by_if_index(
                        IP4, if_index
                    )

                async def set_ip6_gw():
                    by_guid_index[guid]["gws"][IP6] = await get_default_gw_by_if_index(
                        IP6, if_index
                    )

                # Execute the above tasks.
                sub_tasks = [set_addr(), set_ip4_gw(), set_ip6_gw()]
                for task in sub_tasks:
                    await task
            except (OSError, asyncio.TimeoutError):
                log_exception()
                return

        if_tasks.append(if_task_func())

    for if_task in if_tasks:
        await if_task

    return by_guid_index


def win_set_gateways(by_guid_index):
    """Returns a netifaces-style gateways dict built from the given GUID-indexed interface data.

    Every NIC that has a configured gateway is listed in ``gws[af]`` regardless
    of whether it is the OS-selected default route. ``gws["default"][af]`` is
    populated only for the NIC the OS picked. The previous version only
    listed NICs that won the default-route slot, which on multi-NIC hosts
    (e.g. a LAN NIC + a mobile NIC) hid working gateways from ``is_nic_default``
    and any caller asking "does this NIC have a gateway?".
    """
    gws = {"default": {}, int(IP4): [], int(IP6): []}

    for _, addr_info in by_guid_index.items():
        addrs = addr_info.get("addr") or {}
        for af in (IP4, IP6):
            gw = addr_info["gws"].get(af)
            # Empty string / None / missing entry == no gateway.
            if not gw:
                # NIC has at least one IP for this AF but no gateway --
                # surfaced loudly because we hit it on Windows XP where
                # ipconfig /all + ping confirmed a working gateway but
                # the introspection API returned blank.  Don't raise --
                # downstream code deals with missing gateways already --
                # but log so the condition is visible.
                if addrs.get(af):
                    log(fstr(
                        "blank gateway -- may be a bug: nic={0} af={1} has IP(s) but no gateway",
                        (addr_info.get("name", "?"), af),
                    ))
                continue

            is_default = af in addr_info["defaults"]
            gws[int(af)].append((gw, addr_info["name"], is_default))

            if is_default:
                gws["default"][int(af)] = (gw, addr_info["name"])

    return gws


class Netifaces:
    """Drop-in Windows replacement for the netifaces module backed by powershell, WMIC, or netsh."""

    AF_INET = IP4
    AF_INET6 = IP6
    AF_LINK = 18

    def __init__(self):
        pass

    async def start(self):
        # Per-version loader capability matrix. Don't invoke backends on
        # Windows versions where they can't work or are known to hang;
        # each fallback carries a real wall-clock cost (CMD_TIMEOUT per
        # subprocess loader = 30 s). Mapping below -- platform.version()
        # returns the NT kernel version, not the marketing version:
        #
        #   NT ver   Marketing               iphlpapi  PS1   WMIC   netsh
        #   5.0      Win 2000 / Server 2000     no      no    no    weak
        #   5.1+     Win XP SP1+                yes     no    yes   yes
        #   5.2      Server 2003                yes     no    yes   yes
        #   6.0      Vista / Server 2008        yes     no    yes   yes
        #   6.1      Win 7 / Server 2008 R2     yes     no    yes   yes
        #   6.2      Win 8 / Server 2012        yes     yes   yes   yes
        #   6.3      Win 8.1 / Server 2012 R2   yes     yes   yes   yes
        #  10.0      Win 10 / 11 / Server 2016+ yes     yes   yes*  yes
        #
        #   iphlpapi.GetAdaptersAddresses     -- XP SP1+. Single ctypes
        #     call, no subprocess. The only loader that surfaces XP's
        #     per-AF v6_no (TCPIP6 has its own ifindex space). We still
        #     guard with iphlpapi_is_supported() in case some locked-
        #     down build can't load the DLL.
        #
        #   PowerShell Get-NetAdapter (NetAdapter module) -- Win 8 / NT
        #     6.2 and later. Doesn't exist on Win 7 / Vista / XP, so the
        #     PS1 loader is gated on NT >= 6.2.
        #
        #   WMIC -- present XP+ but problematic on XP (single-threaded
        #     repository, hangs on contention) and deprecated in Win 11
        #     22H2+ where Microsoft started removing the binary by
        #     default. Skip on XP entirely (iphlpapi covers it); on
        #     newer Windows we keep WMIC as a fallback even though it
        #     may eventually be absent on a stripped-down 11/Server.
        #
        #   netsh -- weak on Win 2000 (some interface verbs missing) and
        #     on XP its IPv4 subcommands depend on a routing service that
        #     ships disabled in some installs. Reliable Vista+. Same
        #     reasoning as WMIC: skip on XP since iphlpapi suffices.
        #
        # Net effect on the matrix:
        #   XP         -> iphlpapi only (one ctypes call, ms-scale)
        #   Vista / 7  -> iphlpapi -> WMIC -> netsh
        #   Win 8 / 8.1 / 10 / 11 / Server 2012+ -> iphlpapi -> PS1 -> WMIC -> netsh
        # Win 2000 sits in the gap (no iphlpapi guarantee, no PS1, weak
        # netsh, no WMIC); we rely on iphlpapi_is_supported() returning
        # False on that ancient OS so we surface a clean failure rather
        # than running shell loaders that won't produce useful output.
        vmaj, vmin, vpatch = [int(x) for x in platform.version().split(".")]

        # Per-loader availability, computed independently from the NT
        # version so each can be reasoned about (and fixed) in
        # isolation. Names map to the kernel version, not marketing.
        nt = (vmaj, vmin)

        # iphlpapi.GetAdaptersAddresses: XP SP1 (NT 5.1+) onwards.
        # We still call iphlpapi_is_supported() because some locked-
        # down installs strip the function; the dynamic check is
        # the source of truth.
        iphlpapi_ok = nt >= (5, 1) and iphlpapi_is_supported()

        # PowerShell NetAdapter cmdlets (Get-NetAdapter): introduced
        # in Win 8 / Server 2012 (NT 6.2). Win 7 / Vista / XP do not
        # ship this module.
        ps1_ok = nt >= (6, 2)

        # WMIC: present XP+ but the XP repository hangs under
        # contention long enough to chew the CMD_TIMEOUT budget;
        # iphlpapi already covers XP, so don't fall back here.
        # Vista / 7 / 8 / 8.1 / 10 / 11 / Server 2012-2022 are fine.
        # Win 11 22H2+ (build >= 22621) ship without WMIC by
        # default; we still try because WMIC may have been re-
        # added by an admin, and the loader fails fast when the
        # binary is absent.
        wmic_ok = nt >= (6, 0)

        # netsh: works since NT 5.1 in principle but the XP "interface
        # ipv4" verbs depend on the Routing and Remote Access service
        # which is disabled by default on workstation SKUs. Skipping on
        # XP avoids the timeout cost. Vista+ runs reliably.
        netsh_ok = nt >= (6, 0)

        vectors = []
        if iphlpapi_ok:
            vectors.append(if_infos_from_iphlpapi)
        if ps1_ok:
            vectors.append(load_ifs_from_ps1)
        if wmic_ok:
            vectors.append(if_infos_from_wmic)
        if netsh_ok:
            vectors.append(if_infos_from_netsh)

        log("[NETIFACES-LOAD] nt={0}.{1} iphlpapi={2} ps1={3} wmic={4} netsh={5}".format(
            vmaj, vmin, iphlpapi_ok, ps1_ok, wmic_ok, netsh_ok,
        ))

        # Try different funcs to load IF info.
        # Retry up to 3 times with backoff: Windows drops shell calls when
        # multiple processes query interfaces simultaneously under parallel load.
        if_infos = []
        for load_attempt in range(3):
            for load_if_info in vectors:
                try:
                    if_infos = await asyncio.wait_for(load_if_info(), CMD_TIMEOUT)
                    if not len(if_infos):
                        continue

                    break
                except (OSError, asyncio.TimeoutError):
                    log_exception()

            if if_infos:
                break

            if load_attempt < 2:
                await asyncio.sleep(2 + load_attempt * 2)

        self.by_guid_index = {}
        for if_info in if_infos:
            self.by_guid_index[if_info["guid"]] = if_info

        # Sanity check.
        if self.by_guid_index == {}:
            raise Exception("Unable to load interfaces.")

        # Setup name index.
        self.by_name_index = {}
        name_counts = {}
        for _, if_info in self.by_guid_index.items():
            # Name used for this interface.
            name = if_info["name"]

            # A duplicate for this if exists.
            if name in name_counts:
                # Increment count for this name.
                name_counts[name] += 1
                count = name_counts[name]

                # Update the name associated with if details.
                name = fstr(
                    "{0} #{1}",
                    (
                        name,
                        count,
                    ),
                )
                if_info["name"] = name
            else:
                # Record first use of this name.
                name_counts[name] = 1

            # Every if_info has a unique name.
            # Because if descriptions aren't unique for the same HW.
            self.by_name_index[name] = if_info

            # Also accept lookups by adapter description (the long
            # driver string like "Intel(R) 82574L Gigabit Network
            # Connection") when the iphlpapi loader surfaces it.
            # WMIC used to return the description string as the
            # "name", so callers passing --nic with the description
            # form continue to match after the switch to iphlpapi-as-
            # primary loader. Skip when description is empty, equals
            # the friendly name, or is already a registered key.
            description = if_info.get("description")
            if description and description != name and description not in self.by_name_index:
                self.by_name_index[description] = if_info

        # Save main gateways used.
        self.gws = win_set_gateways(self.by_guid_index)
        return self

    def gateways(self):
        """Returns the netifaces-style gateways dict for all detected interfaces."""
        return self.gws

    def if_info(self, if_name):
        """Returns the full interface info dict for the given interface name."""
        return self.by_name_index[if_name]

    def guid(self, if_name):
        """Returns the GUID string for the given interface name."""
        if_info = self.if_info(if_name)
        return if_info["guid"]

    def nic_no(self, if_name):
        """Returns the integer interface index for the given interface name."""
        if_info = self.if_info(if_name)
        return if_info["no"]

    def get_nic_id(self, af, if_name):
        """Return the kernel-level interface index for if_name appropriate to af.

        af=None or IP4 -> the v4 ifindex (same as nic_no).
        af=IP6 -> the v6 ifindex; differs from the v4 ifindex on XP
        where TCPIP and TCPIP6 are separate services with separate
        index spaces. Vista+ unified the stacks so both indices match.
        Loaders that didn't capture the v6 index separately (netsh,
        wmic, ps1, win_net) fall back to the single "no" value.
        """
        if_info = self.if_info(if_name)
        if af == IP6:
            return if_info.get("v6_no") or if_info["no"]
        return if_info["no"]

    def ifaddresses(self, if_name):
        """Returns a netifaces-style address dict keyed by address family for the given interface."""
        if_info = self.by_name_index[if_name]
        addr_format = {
            int(IP4): [],
            int(IP6): [],
            # Netifaces AF_LINK = MAC address.
            Netifaces.AF_LINK: [{"addr": if_info["mac"]}],
        }

        # Add addresses in netiface format.
        for af in [IP4, IP6]:
            for addr_info in if_info["addr"][af]:
                addr = {
                    "addr": addr_info["addr"],
                    "netmask": cidr_to_netmask(addr_info["host_limit"], af),
                }

                addr_format[int(af)].append(addr)

        return addr_format

    def interfaces(self):
        """Returns a sorted list of all detected interface names.

        ONE entry per physical NIC.  The dual-name indexing (friendly
        name AND adapter description) only applies to lookup -- see
        by_name_index registration at the top of __init__ -- because
        we want --nic to match either string.  Enumeration must NOT
        emit both, otherwise load_interfaces() instantiates a fresh
        Interface and runs the full NAT-LOAD pipeline on the SAME
        physical NIC twice, costing 4-10 s of startup on every
        Windows host with adapters whose friendly name differs from
        their description (i.e. all of them on modern Windows).

        Canonical name is the friendly name when present, falling
        back to description.  This matches what the user typed in
        Network Connections and what historical warpgate code expected.

        Loopback adapters (Windows' "Loopback Pseudo-Interface 1" and
        similar) are filtered here.  They still live in by_name_index
        so explicit lookups work, but they MUST NOT appear in default
        enumeration -- callers run the full IFLOAD/STUN/NAT pipeline on
        every returned name, and a loopback-bound STUN probe times out
        4 s per server because the kernel refuses to route loopback-
        sourced packets to a public destination.  test_all_phases
        cold-start surfaced this on Vista: every outbound dial was
        bound to 127.0.0.1 and silently timed out.
        """
        ifs = []
        seen = set()
        for _, if_info in self.by_guid_index.items():
            canonical = if_info.get("name") or if_info.get("description")
            if not canonical or canonical in seen:
                continue
            if is_loopback_iface(if_info):
                log(fstr(
                    "win_netifaces.interfaces: skipping loopback adapter "
                    "name={0}",
                    (canonical,),
                ))
                continue
            ifs.append(canonical)
            seen.add(canonical)

        ifs = sorted(ifs)
        return ifs


async def workspace():
    netifaces = Netifaces()
    await netifaces.start()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(workspace())



from aionetiface import *


async def example():
    # Load default interface.
    nic = await Interface()

    # Get a route to use for sockets.
    # This will give you a copy of the first route for that address family.
    # Routes belong to an interface and include a reference to it.
    route = await nic.route(IP4).bind()

    # Lookup Google.com's IP address -- specify a specific address family.
    # Most websites support IPv4 but not always IPv6.
    # Interface is needed to resolve some specialty edge-cases.
    dest = ("example.com", 80)

    # Now open a TCP connection to that the destination.
    pipe = await Pipe(TCP, dest, route).connect()

    # Send it a malformed HTTP request.
    buf = b"Test\r\n\r\n"
    await pipe.send(buf)

    # Wait for any message response.
    out = await pipe.recv(timeout=3)
    print(out)

    # Cleanup.
    await pipe.close()


if __name__ == "__main__":
    async_test(example)



from abc import ABC
from abc import abstractmethod
import errno
import os


class _InterProcessReaderWriterLockMechanism(ABC):
    @staticmethod
    @abstractmethod
    def trylock(lockfile, exclusive): ...

    @staticmethod
    @abstractmethod
    def unlock(lockfile): ...

    @staticmethod
    @abstractmethod
    def get_handle(path): ...

    @staticmethod
    @abstractmethod
    def close_handle(lockfile): ...


class _InterProcessMechanism(ABC):
    @staticmethod
    @abstractmethod
    def trylock(lockfile): ...

    @staticmethod
    @abstractmethod
    def unlock(lockfile): ...


class _WindowsInterProcessMechanism(_InterProcessMechanism):
    """Interprocess lock implementation that works on windows systems."""

    @staticmethod
    def trylock(lockfile):
        fileno = lockfile.fileno()
        msvcrt.locking(fileno, msvcrt.LK_NBLCK, 1)

    @staticmethod
    def unlock(lockfile):
        fileno = lockfile.fileno()
        msvcrt.locking(fileno, msvcrt.LK_UNLCK, 1)


class _FcntlInterProcessMechanism(_InterProcessMechanism):
    """Interprocess lock implementation that works on posix systems."""

    @staticmethod
    def trylock(lockfile):
        fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)

    @staticmethod
    def unlock(lockfile):
        fcntl.lockf(lockfile, fcntl.LOCK_UN)


class _WindowsInterProcessReaderWriterLockMechanism(
    _InterProcessReaderWriterLockMechanism
):
    """Interprocess readers writer lock implementation that works on windows
    systems."""

    @staticmethod
    def trylock(lockfile, exclusive):

        if exclusive:
            flags = LOCKFILE_FAIL_IMMEDIATELY | LOCKFILE_EXCLUSIVE_LOCK
        else:
            flags = LOCKFILE_FAIL_IMMEDIATELY

        handle = msvcrt.get_osfhandle(lockfile.fileno())
        ok = LockFileEx(handle, flags, 0, 1, 0, pointer(OVERLAPPED()))
        if ok:
            return True
        else:
            last_error = GetLastError()
            if last_error == ERROR_LOCK_VIOLATION:
                return False
            else:
                raise OSError(last_error)

    @staticmethod
    def unlock(lockfile):
        handle = msvcrt.get_osfhandle(lockfile.fileno())
        ok = UnlockFileEx(handle, 0, 1, 0, pointer(OVERLAPPED()))
        if not ok:
            raise OSError(GetLastError())

    @staticmethod
    def get_handle(path):
        return open(path, "a+")

    @staticmethod
    def close_handle(lockfile):
        lockfile.close()


class _FcntlInterProcessReaderWriterLockMechanism(
    _InterProcessReaderWriterLockMechanism
):
    """Interprocess readers writer lock implementation that works on posix
    systems."""

    @staticmethod
    def trylock(lockfile, exclusive):

        if exclusive:
            flags = fcntl.LOCK_EX | fcntl.LOCK_NB
        else:
            flags = fcntl.LOCK_SH | fcntl.LOCK_NB

        try:
            fcntl.lockf(lockfile, flags)
            return True
        except (IOError, OSError) as e:
            if e.errno in (errno.EACCES, errno.EAGAIN):
                return False
            else:
                raise e

    @staticmethod
    def unlock(lockfile):
        fcntl.lockf(lockfile, fcntl.LOCK_UN)

    @staticmethod
    def get_handle(path):
        return open(path, "a+")

    @staticmethod
    def close_handle(lockfile):
        lockfile.close()


if os.name == "nt":
    import msvcrt
    from .pywintypes import OVERLAPPED
    from .win32con import LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY
    from .pywin32.win32file import GetLastError, pointer, UnlockFileEx
    from .pywin32.win32file import ERROR_LOCK_VIOLATION, LockFileEx

    _interprocess_reader_writer_mechanism = (
        _WindowsInterProcessReaderWriterLockMechanism()
    )
    _interprocess_mechanism = _WindowsInterProcessMechanism()

else:
    import fcntl

    _interprocess_reader_writer_mechanism = (
        _FcntlInterProcessReaderWriterLockMechanism()
    )
    _interprocess_mechanism = _FcntlInterProcessMechanism()



from aionetiface import *


async def example():
    netifaces = await aionetiface_setup_netifaces()
    ifs = netifaces.interfaces()
    # e.g. ['lo0', 'en0']

    info = netifaces.ifaddresses("ens192")
    # {18: [{'addr': '8c:...'}], 30: [{'addr': 'fe80::...%en0',
    # 'netmask': 'ffff:ffff:ffff:ffff::/64', 'flags': 1024},
    # {'addr': 'fdf4:1...', 'netmask': 'ffff:ffff:ffff:ffff::/
    # 64', 'flags': 1088}], 2: [{'addr': '192.168.21.144',
    # 'netmask': '255.255.255.0', 'broadcast': '...'}]}

    gws = netifaces.gateways()
    # {'default': {2: ('192.168.21.1', 'en0')},
    # 2: [('192.168.21.1', 'en0', True)],
    # 30: [...]}


if __name__ == "__main__":
    async_test(example)



# -*- coding: utf-8 -*-

# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager
import errno
import functools
import logging
import os
from pathlib import Path
import threading
import time

from ._utils import *
from .process_mechanism import _interprocess_mechanism
from .process_mechanism import _interprocess_reader_writer_mechanism

LOG = logging.getLogger(__name__)


def _ensure_tree(path):
    """Create a directory (and any ancestor directories required).

    :param path: Directory to create
    """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            if not os.path.isdir(path):
                raise
            else:
                return False
        elif e.errno == errno.EISDIR:
            return False
        else:
            raise
    else:
        return True


class InterProcessLock:
    """An interprocess lock."""

    MAX_DELAY = 0.1  # For backwards compatibility
    DELAY_INCREMENT = 0.01  # For backwards compatibility

    def __init__(
        self,
        path,
        sleep_func=time.sleep,
        logger=None,
    ):
        """
        args:
            path:
                Path to the file that will be used for locking.
            sleep_func:
                Optional function to use for sleeping.
            logger:
                Optional logger to use for logging.
        """
        self.lockfile = None
        self.path = canonicalize_path(path)
        self.acquired = False
        self.sleep_func = sleep_func
        self.logger = pick_first_not_none(logger, LOG)

    def __del__(self):
        """Ensure the file handle is closed when the object is garbage collected."""
        try:
            if self.lockfile is not None:
                self._do_close()
        except Exception:
            # Suppress errors during object destruction (e.g. interpreter shutdown)
            pass

    def _try_acquire(self, blocking, watch):
        try:
            self.trylock()
        except (IOError, OSError) as e:
            if e.errno in (errno.EACCES, errno.EAGAIN):
                if not blocking or watch.expired():
                    return False
                else:
                    raise RetryAgain()
            else:
                raise threading.ThreadError(
                    "Unable to acquire lock on"
                    " `%(path)s` due to"
                    " %(exception)s"
                    % {
                        "path": self.path,
                        "exception": e,
                    }
                )
        else:
            return True

    def _do_open(self):
        basedir = os.path.dirname(self.path)
        if basedir:
            made_basedir = _ensure_tree(basedir)
            if made_basedir:
                self.logger.log(BLATHER, "Created lock base path `%s`", basedir)
        # Open in append mode so we don't overwrite any potential contents of
        # the target file. This eliminates the possibility of an attacker
        # creating a symlink to an important file in our lock path.
        if self.lockfile is None or self.lockfile.closed:
            self.lockfile = open(self.path, "a")

    def acquire(
        self,
        blocking=True,
        delay=0.01,
        max_delay=0.1,
        timeout=None,
    ):
        """Attempt to acquire the lock.

        Args:
            blocking:
                Whether to wait to try to acquire the lock.
            delay:
                When `blocking`, starting delay as well as the delay increment
                (in seconds).
            max_delay:
                When `blocking` the maximum delay in between attempts to
                acquire (in seconds).
            timeout:
                When `blocking`, maximal waiting time (in seconds).

        Returns:
            whether or not the acquisition succeeded
        """
        if delay < 0:
            raise ValueError("Delay must be greater than or equal to zero")
        if timeout is not None and timeout < 0:
            raise ValueError("Timeout must be greater than or equal to zero")
        if delay >= max_delay:
            max_delay = delay

        try:
            self._do_open()
            watch = StopWatch(duration=timeout)
            r = Retry(delay, max_delay, sleep_func=self.sleep_func, watch=watch)
            with watch:
                gotten = r(self._try_acquire, blocking, watch)

            if not gotten:
                self.acquired = False
                return False
            else:
                self.acquired = True
                self.logger.log(
                    BLATHER,
                    "Acquired file lock `%s` after waiting %0.3fs [%s"
                    " attempts were required]",
                    self.path,
                    watch.elapsed(),
                    r.attempts,
                )
                return True
        finally:
            # Crucial fix: If we failed to acquire (gotten is False, or exception raised),
            # we MUST close the file handle immediately.
            if not self.acquired:
                self._do_close()

    def _do_close(self):
        if self.lockfile is not None:
            try:
                self.lockfile.close()
            except (IOError, OSError):
                pass
            finally:
                self.lockfile = None

    def __enter__(self):
        gotten = self.acquire()
        if not gotten:
            # This shouldn't happen, but just in case...
            raise threading.ThreadError(
                "Unable to acquire a file lock"
                " on `%s` (when used as a"
                " context manager)" % self.path
            )
        return self

    def release(self):
        """Release the previously acquired lock."""
        if not self.acquired:
            raise threading.ThreadError("Unable to release an unacquired lock")
        try:
            self.unlock()
        except Exception as e:
            msg = "Could not unlock the acquired lock opened on `%s`", self.path
            self.logger.exception(msg)
            raise threading.ThreadError(msg) from e
        else:
            self.acquired = False
            try:
                self._do_close()
            except IOError:
                self.logger.exception(
                    "Could not close the file handle opened on `%s`", self.path
                )
            else:
                self.logger.log(
                    BLATHER, "Unlocked and closed file lock open on `%s`", self.path
                )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def exists(self):
        return os.path.exists(self.path)

    def trylock(self):
        _interprocess_mechanism.trylock(self.lockfile)

    def unlock(self):
        _interprocess_mechanism.unlock(self.lockfile)


class InterProcessReaderWriterLock:
    """An interprocess readers writer lock."""

    MAX_DELAY = 0.1  # for backwards compatibility
    DELAY_INCREMENT = 0.01  # for backwards compatibility

    def __init__(
        self,
        path,
        sleep_func=time.sleep,
        logger=None,
    ):
        """
        Args:
            path:
                Path to the file that will be used for locking.
            sleep_func:
                Optional function to use for sleeping.
            logger:
                Optional logger to use for logging.
        """
        self.lockfile = None
        self.path = canonicalize_path(path)
        self.sleep_func = sleep_func
        self.logger = pick_first_not_none(logger, LOG)

    def __del__(self):
        """Ensure the file handle is closed when the object is garbage collected."""
        try:
            if self.lockfile is not None:
                self._do_close()
        except Exception:
            pass

    @contextmanager
    def read_lock(self, delay=0.01, max_delay=0.1):
        """Context manager that grans a read lock"""

        self.acquire_read_lock(
            blocking=True, delay=delay, max_delay=max_delay, timeout=None
        )
        try:
            yield
        finally:
            self.release_read_lock()

    @contextmanager
    def write_lock(self, delay=0.01, max_delay=0.1):
        """Context manager that grans a write lock"""

        gotten = self.acquire_write_lock(
            blocking=True, delay=delay, max_delay=max_delay, timeout=None
        )

        if not gotten:
            # This shouldn't happen, but just in case...
            raise threading.ThreadError(
                "Unable to acquire a file lock"
                " on `%s` (when used as a"
                " context manager)" % self.path
            )
        try:
            yield
        finally:
            self.release_write_lock()

    def _try_acquire(self, blocking, watch, exclusive):
        try:
            gotten = _interprocess_reader_writer_mechanism.trylock(
                self.lockfile, exclusive
            )
        except Exception as e:
            raise threading.ThreadError(
                "Unable to acquire lock on {} due to {}!".format(self.path, e)
            )

        if gotten:
            return True

        if not blocking or watch.expired():
            return False

        raise RetryAgain()

    def _do_open(self):
        basedir = os.path.dirname(self.path)
        if basedir:
            made_basedir = _ensure_tree(basedir)
            if made_basedir:
                self.logger.log(BLATHER, "Created lock base path `%s`", basedir)
        if self.lockfile is None:
            self.lockfile = _interprocess_reader_writer_mechanism.get_handle(self.path)

    def acquire_read_lock(
        self,
        blocking=True,
        delay=0.01,
        max_delay=0.1,
        timeout=None,
    ):
        """Attempt to acquire a reader's lock."""
        return self._acquire(blocking, delay, max_delay, timeout, exclusive=False)

    def acquire_write_lock(
        self,
        blocking=True,
        delay=0.01,
        max_delay=0.1,
        timeout=None,
    ):
        """Attempt to acquire a writer's lock."""
        return self._acquire(blocking, delay, max_delay, timeout, exclusive=True)

    def _acquire(
        self, blocking=True, delay=0.01, max_delay=0.1, timeout=None, exclusive=True
    ):
        if delay < 0:
            raise ValueError("Delay must be greater than or equal to zero")
        if timeout is not None and timeout < 0:
            raise ValueError("Timeout must be greater than or equal to zero")
        if delay >= max_delay:
            max_delay = delay

        success = False
        try:
            self._do_open()
            watch = StopWatch(duration=timeout)
            r = Retry(delay, max_delay, sleep_func=self.sleep_func, watch=watch)
            with watch:
                gotten = r(self._try_acquire, blocking, watch, exclusive)

            if gotten:
                self.logger.log(
                    BLATHER,
                    "Acquired file lock `%s` after waiting %0.3fs [%s"
                    " attempts were required]",
                    self.path,
                    watch.elapsed(),
                    r.attempts,
                )
                success = True
                return True
            else:
                return False
        finally:
            if not success:
                self._do_close()

    def _do_close(self):
        if self.lockfile is not None:
            _interprocess_reader_writer_mechanism.close_handle(self.lockfile)
            self.lockfile = None

    def release_write_lock(self):
        """Release the writer's lock."""
        try:
            _interprocess_reader_writer_mechanism.unlock(self.lockfile)
        except IOError:
            self.logger.exception(
                "Could not unlock the acquired lock opened on `%s`", self.path
            )
        else:
            try:
                self._do_close()
            except IOError:
                self.logger.exception(
                    "Could not close the file handle opened on `%s`", self.path
                )
            else:
                self.logger.log(
                    BLATHER, "Unlocked and closed file lock open on `%s`", self.path
                )

    def release_read_lock(self):
        """Release the reader's lock."""
        try:
            _interprocess_reader_writer_mechanism.unlock(self.lockfile)
        except IOError:
            self.logger.exception(
                "Could not unlock the acquired lock opened on `%s`", self.path
            )
        else:
            try:
                self._do_close()
            except IOError:
                self.logger.exception(
                    "Could not close the file handle opened on `%s`", self.path
                )
            else:
                self.logger.log(
                    BLATHER, "Unlocked and closed file lock open on `%s`", self.path
                )


def interprocess_write_locked(path):
    """Acquires & releases an interprocess  **write** lock around the call into
    the decorated function

    Args:
        path: Path to the file used for locking.
    """
    lock = InterProcessReaderWriterLock(path)

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with lock.write_lock():
                return f(*args, **kwargs)

        return wrapper

    return decorator


def interprocess_read_locked(path):
    """Acquires & releases an interprocess **read** lock around the call into
    the decorated function

    Args:
        path: Path to the file used for locking.
    """
    lock = InterProcessReaderWriterLock(path)

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with lock.read_lock():
                return f(*args, **kwargs)

        return wrapper

    return decorator


def interprocess_locked(path):
    """Acquires & releases an interprocess lock around the call to the
    decorated function.

    Args:
        path: Path to the file used for locking.
    """
    lock = InterProcessLock(path)

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with lock:
                return f(*args, **kwargs)

        return wrapper

    return decorator



import binascii
from aionetiface import *


async def example():
    # Open a UDP pipe to google's STUN server.
    pipe = await Pipe(UDP, ("stun.l.google.com", 19302)).connect()

    # Random STUN message ID.
    msg_id = binascii.hexlify(rand_b(12))

    #  req  = req type    len     magic cookie
    req_hex = b"0001" + b"0000" + b"2112A442" + msg_id
    req_buf = binascii.unhexlify(req_hex)

    # UDP is unreliable -- try up to 3 times.
    for _ in range(0, 3):
        # Send STUN bind request and get resp.
        await pipe.send(req_buf)
        resp = await pipe.recv()

        # Timeout -- try again.
        if resp is None:
            continue

        # Show resp -- exit loop.
        print(resp)
        break

    # Cleanup.
    await pipe.close()


if __name__ == "__main__":
    async_test(example)



from aionetiface import *
from aionetiface.testing import AsyncTestCase
from port_helpers import xdist_port_base


class TestRUDP(AsyncTestCase):
    async def test_rudp(self):
        i = await Interface()
        af = i.supported()[0]
        port = xdist_port_base(40000)
        r = await i.route(af).bind(port)
        dest_tup = (r.nic(), port)
        dest = dest_tup
        pipe = await Pipe(route=r, proto=RUDP, dest=dest).connect()

        msg = b"test meow"
        task, event = await pipe.stream.ack_send(msg, dest_tup)
        await asyncio.wait_for(event.wait(), 2)
        out = await pipe.recv(SUB_ALL)
        self.assertEqual(out, msg)
        await pipe.close()

    async def test_is_unique(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        b = PipeEvents(sock=s)
        d = b"ayy lmao"
        c = ("127.0.0.1", 1337)
        r = b.is_unique_msg(pipe=None, data=d, client_tup=c)

        self.assertTrue(r)
        s.close()


if __name__ == "__main__":
    main()



"""AFGroup -- per-AF Interface selection.

Most clients (Router, MQTT, SmartPipe, Address, ...) historically took
a single Interface and assumed it was workable for both v4 and v6.
That assumption breaks the moment a host has different NICs handling
different AFs:

  * Mobile dongle is v4-only CGNAT (the actual NAT-traversal target);
    primary NIC is dual-stack -- v4 brokers must go via mobile, v6
    brokers via primary.
  * Test VMs with v6 disabled on one NIC but enabled on another.
  * Multi-homed hosts with management-only v4 and services-only v6.

AFGroup is a thin wrapper around {af: Interface} that

  * accepts a single dual-stack Interface and fans it across the AFs
    that interface supports (legacy compat -- pass a single iface,
    everything still works);
  * accepts an explicit dict so multi-homed callers can split AFs
    across NICs;
  * validates per-AF support at construction so a "this NIC doesn't
    actually have v6" mistake is caught at the boundary instead of as
    a NoneType error three layers deep;
  * exposes for_af / supports / iteration so consuming code never
    reaches into a raw dict.
"""

from ..utility.utils import fstr
from .interface import Interface


class AFGroup:
    """The set of Interfaces a client uses, keyed by address family."""

    def __init__(
        self,
        interfaces,
    ):
        """Build the group from another AFGroup, a {af: Interface} dict, or a single Interface."""
        by_af = {}
        if isinstance(interfaces, AFGroup):
            by_af = dict(interfaces.by_af)
        elif isinstance(interfaces, dict):
            for af, iface in interfaces.items():
                if iface is None:
                    raise ValueError(
                        fstr("AFGroup: interface for af {0} is None", (af,))
                    )
                if af not in iface.what_afs():
                    raise ValueError(fstr(
                        "AFGroup: interface {0} does not support af {1}",
                        (repr(iface.name), af),
                    ))
                by_af[af] = iface
        elif isinstance(interfaces, Interface):
            for af in interfaces.what_afs():
                by_af[af] = interfaces
        else:
            raise TypeError(fstr(
                "AFGroup: expected Interface, dict, or AFGroup; got {0}",
                (repr(type(interfaces).__name__),),
            ))

        self.by_af = by_af

    def for_af(self, af):
        """Return the Interface assigned to *af*; raise KeyError if none."""
        if af not in self.by_af:
            raise KeyError(
                fstr("AFGroup: no interface configured for af {0}", (af,))
            )
        return self.by_af[af]

    def get(self, af, default=None):
        """Return the Interface for *af*, or *default* when none is configured."""
        return self.by_af.get(af, default)

    def supports(self, af):
        """True iff this group has an Interface configured for *af*."""
        return af in self.by_af

    def afs(self):
        """Return the address families this group covers, in insertion order."""
        return list(self.by_af.keys())

    # ------------------------------------------------------------------
    # Interface-mimic API
    #
    # These mirror the per-AF accessors on Interface so AFGroup is a
    # drop-in replacement in callers that previously took an Interface.
    # The Interface methods that take an *implicit* primary AF (zero-arg
    # `route()` / `nic()`) are intentionally NOT mirrored -- AFGroup has
    # no single primary by definition; callers that used those forms
    # need to pass an explicit `af`. The compiler / runtime will catch
    # the missing-arg call sites cleanly.
    # ------------------------------------------------------------------

    def route(self, af, bind_port=0):
        """Return the primary Route on the Interface chosen for *af*."""
        return self.for_af(af).route(af, bind_port)

    def nic(self, af):
        """Return the NIC IP string for the Interface chosen for *af*."""
        return self.for_af(af).nic(af)

    def supported(self, skip_resolve=0):
        """Return the AFs this group covers (matches Interface.supported's signature)."""
        return self.afs()

    def what_afs(self):
        """Alias for supported(); matches Interface.what_afs."""
        return self.afs()

    def is_default(self, af, gws=None):
        """Return whether the Interface chosen for *af* is the host's default route."""
        return self.for_af(af).is_default(af, gws)

    @classmethod
    def from_interfaces(cls, interfaces):
        """Build an AFGroup from a list of Interfaces, picking first-match per AF.

        Walks *interfaces* in order; for each AF, the first interface that
        supports it wins. Lets a caller express simple per-AF preferences
        purely by NIC ordering -- e.g. pass [mobile_v4_only, primary_dual]
        and the v4 path uses the mobile NIC while the v6 path uses primary.
        Callers that want a different mapping (or to skip an AF entirely)
        should construct AFGroup with an explicit dict instead.
        """
        if not interfaces:
            raise ValueError("AFGroup.from_interfaces: empty interfaces list")
        by_af = {}
        for iface in interfaces:
            for af in iface.what_afs():
                if af not in by_af:
                    by_af[af] = iface
        if not by_af:
            raise ValueError("AFGroup.from_interfaces: no AF coverage across given interfaces")
        return cls(by_af)

    def interfaces(self):
        """Return the unique Interfaces in this group (deduped, insertion order)."""
        seen = []
        for iface in self.by_af.values():
            if iface not in seen:
                seen.append(iface)
        return seen

    def __iter__(self):
        return iter(self.by_af.items())

    def __len__(self):
        return len(self.by_af)

    def __contains__(self, af):
        return af in self.by_af

    def __repr__(self):
        body = ", ".join(
            "{0}: {1!r}".format(af, iface.name)
            for af, iface in self.by_af.items()
        )
        return "AFGroup({" + body + "})"



"""High-level route-table queries (default gateway, internet check)."""
import re
import platform
from ...utility.utils import async_test, ip_f
from ...utility.cmd_tools import cmd
from ...net.net_defs import IP4, IP6, VALID_AFS
from ...net.net_utils import ip_norm


async def windows_get_route_table(af):
    table = []
    if af == IP4:
        cmd_buf = 'powershell "Get-NetRoute -AddressFamily IPv4"'
    if af == IP6:
        cmd_buf = 'powershell "Get-NetRoute -AddressFamily IPv6"'

    out = await cmd(cmd_buf, timeout=4)
    p = r"([0-9]+)\s+([^\s]+)\s+([^\s]+)\s+([0-9]+)\s+([0-9]+)\s+([^\s]+)[\r\n]*"
    results = re.findall(p, out)
    for result in results:
        entry = {
            "if_index": int(result[0]),
            "if": int(result[0]),
            "dest": result[1],
            "next_hop": result[2],
            "route_metric": int(result[3]),
            "if_metric": int(result[4]),
            "policy_store": result[5],
        }

        table.append(entry)

    return table


async def linux_get_route_table(af):
    table = []
    bin_path = "/usr/sbin/route"
    if af == IP4:
        cmd_buf = "{} -4".format(bin_path)
        out = await cmd(cmd_buf, timeout=4)
        p = r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([-0-9]+)\s+([-0-9]+)\s+([-0-9]+)\s+([^\r\n]+)[\r\n]*"
        results = re.findall(p, out)
        for result in results:
            entry = {
                "dest": result[0],
                "gw": result[1],
                "gen_mask": result[2],
                "flag": result[3],
                "metric": int(result[4]),
                "ref": int(result[5]),
                "use": int(result[6]),
                "if": result[7],
            }

            table.append(entry)

    if af == IP6:
        cmd_buf = "{} -6".format(bin_path)
        out = await cmd(cmd_buf, timeout=4)
        p = r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([-0-9]+)\s+([-0-9]+)\s+([-0-9]+)\s+([^\r\n]+)[\r\n]*"
        results = re.findall(p, out)
        for result in results:
            entry = {
                "dest": result[0],
                "next_hop": result[1],
                "flag": result[2],
                "metric": int(result[3]),
                "ref": int(result[4]),
                "use": int(result[5]),
                "if": result[6],
            }

            table.append(entry)

    return table


async def darwin_get_route_table(af):
    table = []
    p = r"([^\r\n]+?)[ ]+([^\r\n]+?)[ ]+([^\r\n]+?)[ ]+([^\r\n ]+)[ ]*(([^\r\n]+)[ ]*)?[\r\n]+"
    if af == IP4:
        cmd_buf = "netstat -rn -f inet"
    if af == IP6:
        cmd_buf = "netstat -rn -f inet6"

    out = await cmd(cmd_buf, timeout=4)
    results = re.findall(p, out)
    for result in results:
        entry = {
            "dest": result[0],
            "gw": result[1],
            "flags": result[2],
            "if": result[3],
            "expiry": result[4],
        }

        table.append(entry)

    return table


async def get_route_table(af):
    if platform.system() == "Windows":
        return await windows_get_route_table(af)

    if platform.system() == "Linux":
        return await linux_get_route_table(af)

    if platform.system() in ["Darwin", "FreeBSD"]:
        return await darwin_get_route_table(af)

    return []


def find_rt_entry(
    dest, if_name, table
):
    """Return the first route-table entry matching dest and if_name, or None."""
    for entry in table:
        if entry["dest"] != dest:
            continue

        if entry["if"] != if_name:
            continue

        return entry


async def darwin_is_internet_if(if_name):
    def is_internet_if(table):
        # Find default entry for iface.
        default_entry = find_rt_entry("default", if_name, table)
        if default_entry is None:
            return False

        # Check flags for entry.
        if "U" not in default_entry["flags"]:
            return False
        if "G" not in default_entry["flags"]:
            return False

        return True

    for af in VALID_AFS:
        table = await get_route_table(af)
        if is_internet_if(table):
            return True

    return False


async def linux_is_internet_if(if_name):
    def is_internet_if(table, af):
        if af == IP6:
            for entry in table:
                if entry["if"] != if_name:
                    continue

                cidr = int(entry["dest"].split("/")[-1])
                if cidr != 128:
                    continue

                if entry["next_hop"] != "[::]":
                    continue

                try:
                    ip_s = ip_norm(entry["dest"])
                except ValueError:
                    continue

                ip_obj = ip_f(ip_s)
                if ip_obj.is_private:
                    continue

                return True

        if af == IP4:
            dest = "default"
            entry = find_rt_entry(dest, if_name, table)
            if entry is None:
                return False

            if entry["gen_mask"] == "0.0.0.0":
                return True

        return False

    for af in VALID_AFS:
        table = await get_route_table(af)
        if is_internet_if(table, af):
            return True

    return False


async def windows_is_internet_if(if_index):
    def is_internet_if(table, af):
        if af == IP4:
            dest = "0.0.0.0/0"
        if af == IP6:
            dest = "::/0"  # Probably what it will be.

        return find_rt_entry(dest, if_index, table) is not None

    for af in VALID_AFS:
        table = await get_route_table(af)
        if is_internet_if(table, af):
            return True

    return False


async def is_internet_if(if_name):
    if platform.system() == "Linux":
        return await linux_is_internet_if(if_name)

    if platform.system() in ["Darwin", "FreeBSD"]:
        return await darwin_is_internet_if(if_name)

    if platform.system() == "Windows":
        return await windows_is_internet_if(if_name)

    return False


if __name__ == "__main__":  # pragma: no cover
    from .interface import Interface

    async def test_route_table():
        i = Interface()
        if i.nic_no:
            await is_internet_if(i.nic_no)
        else:
            await is_internet_if(i.name)

    async_test(test_route_table)



"""Factory helpers for building concurrent async patterns."""
import asyncio
from collections import defaultdict
from .utils import fstr, log


async def concurrent_first_agree_or_best(
    min_agree, tasks, timeout, wait_all=False
):
    """Run tasks concurrently and return the first result agreed upon by min_agree tasks."""
    results = defaultdict(int)
    log(
        fstr(
            "Con first agree: min={0}. task_no={1}, timeout={2}, wait_all={3}",
            (min_agree, len(tasks), timeout, wait_all),
        )
    )

    def check_consensus(result):
        """Return result if it has now reached the min_agree threshold, else None."""
        if result is None or isinstance(result, Exception):
            return None
        results[result] += 1
        if results[result] >= min_agree:
            return result
        return None

    # Convert coroutines to Tasks so they can be individually cancelled.
    scheduled = [
        t if isinstance(t, asyncio.Task) else asyncio.ensure_future(t) for t in tasks
    ]

    winner = None
    try:
        if wait_all:
            done, pending = await asyncio.wait(scheduled, timeout=timeout)
            for t in pending:
                t.cancel()
            for t in done:
                try:
                    winner = check_consensus(t.result())
                except (Exception, asyncio.CancelledError):
                    pass
                if winner is not None:
                    break
        else:
            try:
                for fut in asyncio.as_completed(scheduled, timeout=timeout):
                    try:
                        result = await fut
                        winner = check_consensus(result)
                    except (Exception, asyncio.CancelledError):
                        pass
                    if winner is not None:
                        break
            except asyncio.TimeoutError:
                pass
    finally:
        for t in scheduled:
            if not t.done():
                t.cancel()

    return winner


async def repeat_every(n, coro_func, *args, **kwargs):
    """Call coro_func with args/kwargs every n seconds indefinitely."""
    while True:
        await coro_func(*args, **kwargs)
        await asyncio.sleep(n)



"""Windows network interface enumeration via netsh output parsing."""
import re
import sys

if sys.platform == "win32":
    import winreg
from ....net.net_defs import IP4, IP6
from ....net.net_utils import ip_strip_if, toggle_host_bits
from ....net.ip_range import IPRange
from ....utility.cmd_tools import cmd
from ....utility.utils import fstr, safe_gather


class NetshParse:
    """Parses netsh command output into structured network interface data."""

    # netsh interface ipv4 show interfaces
    # netsh interface ipv6 show interfaces
    # if_index:
    @staticmethod
    def show_interfaces(af, msg):
        """Returns a list of parsed interface records indexed by interface index."""
        p = r"([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([a-z0-9]+)\s+([^\r\n]+)"
        out = re.findall(p, msg)
        results = {}
        for match_group in out:
            if_index, metric, mtu, state, name = match_group
            if if_index not in results:
                results[if_index] = {
                    "if_index": if_index,
                    "metric": metric,
                    "mtu": mtu,
                    "state": state,
                    "con_name": name,
                }

        return [af, "ifs", results]

    # netsh interface ipv4 show ipaddresses
    # netsh interface ipv6 show addresses
    # if_index: ...
    @staticmethod
    def show_addresses(af, msg):
        """Returns a list of parsed address records grouped by interface index."""
        msg = re.sub(r"%[0-9]+", "", msg)

        # Regex patterns that can match address information.
        # The pattern looks for the start of the interface line.
        # To tether it so it doesn't just match all text.
        p = r"[Ii]nterface\s+([0-9]+)[\s\S]+?[\r\n]([a-zA-Z0-9]+)\s+([a-zA-Z0-9]+)\s+([a-zA-Z0-9]+)\s+([a-zA-Z0-9]+)\s+((?=\S*[0-9]+\S*)([a-fA-F0-9:.]+))"

        # Build a table of all address info for each interface.
        # The table is indexed by interface no / if_index.
        results = {}
        while True:
            # Find a valid address line for an interface
            addr_infos = re.findall(p, msg)
            if not len(addr_infos):
                break

            for addr_info in addr_infos:
                # Unpack the result.
                if_index, addr_type, dad_state, valid_life, pref_life, addr = addr_info[
                    :6
                ]
                if if_index not in results:
                    results[if_index] = []

                # Record details as a keyed record.
                results[if_index].append(
                    {
                        "addr_type": addr_type,
                        "dad_state": dad_state,
                        "valid_life": valid_life,
                        "pref_life": pref_life,
                        "addr": addr,
                    }
                )

                # Remove the interface address line from the string.
                # Otherwise the regex will match the same result.
                # print(msg)
                msg = re.sub(p, fstr("Interface {0}\r\n", (if_index,)), msg, count=1)

        return [af, "addrs", results]

    # netsh interface ipv4 show route
    # netsh interface ipv6 show route
    # Routes also show subnet for interface addresses.
    @staticmethod
    def show_route(af, msg):
        """Returns a list of parsed route records grouped by interface index."""
        p = r"([a-zA-Z0-9]+)\s+([a-zA-Z0-9]+)\s+([0-9]+)\s+([a-zA-Z0-9.:%\/]+)\s+([0-9]+)\s+([^\r\n]+)"
        out = re.findall(p, msg)
        results = {}
        for match_group in out:
            publish, rtype, metric, prefix, if_index, con_name = match_group
            if if_index not in results:
                results[if_index] = []

            results[if_index].append(
                {
                    "publish": publish,
                    "rtype": rtype,
                    "metric": metric,
                    "prefix": prefix,
                    "con_name": con_name,
                }
            )

        return [af, "routes", results]

    # route print
    # Also has ipv6 results.
    # if_index: ... if_name, mac
    @staticmethod
    def show_mac(af, msg):
        """Returns a list of parsed MAC address and default gateway records by interface index."""
        p = r"([0-9]+)\s*[.]{2,}([0-9a-fA-F ]+)[ .]+([^\r\n]+)[\r\n]"
        out = re.findall(p, msg)
        results = {"default": {IP4: None, IP6: None}}
        for match_group in out:
            if_index, mac, if_name = match_group
            mac = mac.strip().lower()
            mac = mac.replace(" ", "-")
            results[if_index] = {"if_name": if_name, "mac": mac}

        # Setup entries for default gateways IP4.
        p = r"0[.]0[.]0[.]0\s+0[.]0[.]0[.]0\s+([^\s]+)\s+([^\s]+)\s+[0-9]+"
        out = re.findall(p, msg)
        if len(out):
            gw_ip, if_ip = out[0]
            results["default"][IP4] = {"gw_ip": gw_ip.strip(), "if_ip": if_ip.strip()}

        # Setup entries for default gateways IP6.
        p = r"[0-9]+\s+[0-9]+\s+::\/0\s+([^\s]+)"
        out = re.findall(p, msg)
        if len(out):
            gw_ip = out[0]
            results["default"][IP6] = {"gw_ip": gw_ip.strip()}

        return [af, "macs", results]

    # ipconfig /all
    # mac: {ip4: ..., IP6: ...}
    @staticmethod
    def show_gws(af, msg):
        """Returns a list of parsed gateway addresses keyed by MAC address."""
        p = r"[pP]hysical[ ]+[aA]ddress[^:]+:([^\r\n]+)[\r\n][\s\S]+?[dD]efault[ ]+[gG]ateway[^:]+:((?:\s*[a-fA-F0-9:.%]+[\r\n]?)(?:\s*[a-fA-F0-9:.%]+[\r\n])?)"
        sections = msg.split("\r\n\r\n")

        results = {}
        for section in sections:
            out = re.findall(p, section)
            for match_group in out:
                mac, gws = match_group
                mac = mac.strip().lower()
                gws = gws.strip()
                gws = gws.split()
                af_gws = {IP4: None, IP6: None}

                success = False
                for offset in range(0, len(gws)):
                    try:
                        gw = ip_strip_if(gws[offset])
                        gw_ipr = IPRange(ip=gw)
                        af_gws[gw_ipr.af] = gw
                    except (ValueError, OSError):
                        continue
                    success = True

                if not success:
                    continue

                results[mac] = af_gws

        return [af, "gws", results]


async def do_netsh_cmds():
    parser = NetshParse()
    cmd_vectors = [
        [parser.show_interfaces, {IP4: "interfaces", IP6: "interfaces"}],
        [parser.show_addresses, {IP4: "ipaddresses", IP6: "addresses"}],
        [parser.show_route, {IP4: "route", IP6: "route"}],
        [
            parser.show_mac,
            {
                IP4: "route print",
            },
            False,
        ],
        [
            parser.show_gws,
            {
                IP4: "ipconfig /all",
            },
            False,
        ],
    ]

    async def helper(af, cmd_val, func):
        out = await cmd(cmd_val)
        return func(af, out)

    tasks = []
    for cmd_vector in cmd_vectors:
        for af in [IP4, IP6]:
            if af in cmd_vector[1]:
                show_val = cmd_vector[1][af]
                af_val = "ipv4" if af == IP4 else "ipv6"
                if len(cmd_vector) > 2:
                    cmd_val = show_val
                else:
                    cmd_val = fstr(
                        "netsh interface {0} show {1}",
                        (
                            af_val,
                            show_val,
                        ),
                    )

                tasks.append(helper(af, cmd_val, cmd_vector[0]))

    # Execute all netsh commands concurrently.
    results = await safe_gather(*tasks)

    # Combine results to make them easier to process.
    info = {}
    for result in results:
        af, k, v = result
        if k == "ifs":
            info["ifs"] = v
            continue

        if k not in info:
            info[k] = {}

        info[k][af] = v

    return info


"""
Net name is very specifically not the interface name or
its description. Examples of the net name are 'local area network.'
Examples of an interface name 'Intel ... ethernet v10'.

"""


def win_con_name_lookup():
    """Returns a mapping of connection names to their GUID and interface description from the Windows registry."""
    root_key = winreg.OpenKey(
        winreg.HKEY_LOCAL_MACHINE,
        r"SYSTEM\ControlSet001\Control\Network",
        0,
        winreg.KEY_READ,
    )

    # Recursively crawl all portions looking for the right field.
    # Windows loves to make things easy.
    def recurse_search(root_key, guid=None):
        """Returns a list of connection name and GUID pairs found under the given registry key."""
        results = []
        for sub_offset in range(0, winreg.QueryInfoKey(root_key)[0]):
            sub_name = winreg.EnumKey(root_key, sub_offset)
            sub_key = None
            try:
                sub_key = winreg.OpenKey(root_key, sub_name)
                if sub_name == "Connection":
                    con_name = winreg.QueryValueEx(sub_key, "Name")[0]
                    results.append([con_name, guid])
                else:
                    if re.match(r"{[^{}]+}", sub_name) is None:
                        continue

                    results += recurse_search(sub_key, guid=sub_name)
            except OSError:
                pass
            finally:
                if sub_key is not None:
                    sub_key.Close()

        return results

    # Build list of con_name -> guid mappings.
    results = recurse_search(root_key)
    root_key.Close()

    # Look up interface names.
    root_key = winreg.OpenKey(
        winreg.HKEY_LOCAL_MACHINE,
        r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\NetworkCards",
        0,
        winreg.KEY_READ,
    )

    # con_name -> {"guid": ..., "if_name": ...}
    infos = {}
    for sub_offset in range(0, winreg.QueryInfoKey(root_key)[0]):
        sub_name = winreg.EnumKey(root_key, sub_offset)
        sub_key = None
        found_guid = found_if_name = ""
        try:
            sub_key = winreg.OpenKey(root_key, sub_name)
            found_guid = winreg.QueryValueEx(sub_key, "ServiceName")[0]
            found_if_name = winreg.QueryValueEx(sub_key, "Description")[0]
        except OSError:
            pass
        finally:
            if sub_key is not None:
                sub_key.Close()

        for result in results:
            con_name, saved_guid = result
            if saved_guid == found_guid:
                infos[con_name] = {"guid": saved_guid, "if_name": found_if_name}

    root_key.Close()
    return infos


def get_host_limit_from_route_infos(
    needle_ip, route_infos, af=None
):
    """Returns the most specific CIDR prefix length and netmask matching the given IP from a list of route infos."""
    from ....net.net_utils import af_bitlen
    netmask = None
    host_limit = None
    for route_info in route_infos:
        prefix_ip, prefix_cidr = route_info["prefix"].split("/")
        prefix_cidr = int(prefix_cidr)
        if not prefix_cidr:
            continue

        import ipaddress as _ipa

        _max_bits = 128 if _ipa.ip_address(prefix_ip).version == 6 else 32
        _host_bits = _max_bits - prefix_cidr
        prefix_ipr = IPRange(prefix_ip, bitlen=_host_bits)
        masked_needle_net = toggle_host_bits(prefix_ipr.netmask, needle_ip)
        masked_needle_ipr = IPRange(masked_needle_net, bitlen=_host_bits)

        if prefix_ipr == masked_needle_ipr:
            if host_limit is None or prefix_cidr >= host_limit:
                host_limit = prefix_cidr
                netmask = prefix_ipr.netmask

    if host_limit is None:
        host_limit = af_bitlen(af) if af is not None else 0
    return [host_limit, netmask]


async def if_infos_from_netsh():
    con_table = win_con_name_lookup()
    out = await do_netsh_cmds()

    if_infos = []
    for if_index in out["ifs"]:
        if_info = out["ifs"][if_index]
        con_name = if_info["con_name"]

        addr_info = {IP4: [], IP6: []}
        for af in [IP4, IP6]:
            if if_index not in out["addrs"][af]:
                continue

            for found_addr in out["addrs"][af][if_index]:
                route_infos = out["routes"][af].get(if_index, [])
                host_limit, netmask = get_host_limit_from_route_infos(
                    found_addr["addr"], route_infos, af
                )

                addr = {
                    "addr": found_addr["addr"],
                    "af": af,
                    "host_limit": host_limit,
                    "netmask": netmask,
                }
                addr_info[af].append(addr)

        if con_name not in con_table:
            continue

        if if_index not in out["macs"][IP4]:
            continue

        mac = out["macs"][IP4][if_index]["mac"].rstrip().lower()
        if mac not in out["gws"][IP4]:
            continue
        gws = out["gws"][IP4][mac]

        # Determine if interface isdefault for any AF.
        defaults = []
        for test_af in [IP4, IP6]:
            af_default = out["macs"][IP4]["default"][test_af]
            if af_default is None:
                continue

            af_gw = gws[test_af]
            if af_gw is None:
                continue

            if IPRange(af_gw) == IPRange(af_default["gw_ip"]):
                defaults.append(test_af)

        result = {
            "con_name": con_name,
            "guid": con_table[con_name]["guid"],
            "name": con_table[con_name]["if_name"],
            "no": int(if_index),
            "mac": mac,
            "addr": addr_info,
            "gws": gws,
            "defaults": defaults,
        }

        if_infos.append(result)

    return if_infos



"""
Given a list of ip strings and a list of nics:
    - create a new route table that consists only of those IPs
    - apply these tables in-place only for the NICs where those IPs belong

accepted inputs:
    - v4 WAN ip
        -> route with that WAN ip and the NIC ips to use it
        - limit wan ips to only the defined set of wan ips
    - v4 NIC ip
        -> route with only that nic ip and the wan ip it points to
    - v6 link-local
        - build a list of link locals for that nic
        - set all the wan ip routes to use that list
    - v6 wan ip
        -> route with ext as only that wan ip
            - if link-locals defined for that nic set only to that list
            - otherwise reuse existing link locals
        - limit wan ips to only the defined set of wan ips
"""
from ...utility.utils import find_intersect
from ...net.net_defs import IP4, IP6
from ...net.ip_range import IPR
from .route import Route
from .route_pool import RoutePool


def sort_iprs(ipr_list):
    """Return a (public, private) tuple of IPR lists split from ipr_list."""
    # Sort by public / private.
    pub_iprs = []
    priv_iprs = []
    for ipr in ipr_list:
        if ipr.is_public:
            pub_iprs.append(ipr)

        if ipr.is_private:
            priv_iprs.append(ipr)

    return pub_iprs, priv_iprs


def v6_route_pool_from_ips(ipr_list, nic):
    """Return a list of IPv6 Route objects built from the given IPR list and NIC."""
    # Sort by public / private.
    pub_iprs, priv_iprs = sort_iprs(ipr_list)

    # If no explicit link-locals were requested, inherit from the NIC.
    # Prefer RoutePool-level link_locals (always populated during loading),
    # falling back to the first route's link_locals when the pool is non-empty.
    if not priv_iprs:
        if nic.rp[IP6].link_locals:
            priv_iprs = nic.rp[IP6].link_locals
        elif nic.rp[IP6].wan_hosts:
            priv_iprs = nic.rp[IP6][0].link_locals

    # Build route pool from existing global IPv6 routes.
    routes = []
    for route in nic.rp[IP6]:
        # If public exts are specified limit to only those routes.
        # Yield from route.ext_ips (NOT pub_iprs) so the IPR keeps the
        # NIC's real prefix length (/64 typically). pub_iprs entries
        # are built via IPR(ip) with no netmask, which defaults to
        # bitlen=0 (single-host) — feeding that into Route would
        # serialise the addr with cidr 0 and break the peer-side
        # subnet match. See v4_route_pool_from_ips for the parallel.
        if pub_iprs:
            pub_ipr = list(find_intersect(route.ext_ips, pub_iprs))
            if not pub_ipr:
                continue

            ext_ips = pub_ipr
        else:
            ext_ips = route.ext_ips

        routes.append(Route(IP6, ext_ips, ext_ips, nic))

    # When the caller specified link-locals but the NIC has no global IPv6
    # routes, create a link-local-only route so the address can be encoded.
    # ext_check=0 is required because link-locals are not globally routable.
    if priv_iprs and not routes and not pub_iprs:
        ll_ipr = priv_iprs[0]
        routes.append(Route(IP6, [ll_ipr], [ll_ipr], nic, ext_check=0))

    # Set link-local list.
    for route in routes:
        route.set_link_locals(priv_iprs)

    return routes


def v4_route_pool_from_ips(ipr_list, nic):
    """Return a list of IPv4 Route objects built from the given IPR list and NIC."""
    # Sort by public / private.
    pub_iprs, priv_iprs = sort_iprs(ipr_list)

    # Build route pool.
    routes = []
    for route in nic.rp[IP4]:
        # Use pre-existing route as template.
        # Yield from route.nic_ips (NOT priv_iprs) so the IPR with the
        # NIC's real subnet bitlen survives -- otherwise the narrowed
        # route inherits the listen-IP's default /32 (IPR(ip) with no
        # netmask defaults to bitlen=0 = single host), the address
        # serialises with cidr 0 instead of /24, and the peer's LAN-
        # subnet match fails ("dest not in my LAN" -> tries EXT instead,
        # which collapses when both peers share the WAN IP).
        nic_ipr = list(find_intersect(route.nic_ips, priv_iprs))
        if not nic_ipr:
            continue

        # Same correctness reason as nic_ipr above -- yield the
        # route's IPR (with its real bitlen) when the listen filter
        # matches.
        wan_ipr = list(find_intersect(route.ext_ips, pub_iprs))
        if wan_ipr:
            ext_ips = wan_ipr
        else:
            ext_ips = route.ext_ips

        routes.append(Route(IP4, nic_ipr, ext_ips, nic))

    return routes


def route_pool_from_ips(ip_list, nic):
    """Return an AF-keyed RoutePool dict derived from ip_list filtered against the given NIC."""
    # Classify IPs by version.
    ipr_list = {IP4: [], IP6: []}
    for ip in ip_list:
        ipr = IPR(ip)
        if ipr.af == IP4:
            ipr_list[IP4].append(ipr)

        if ipr.af == IP6:
            ipr_list[IP6].append(ipr)

    # AF-specific route pool funcs.
    get_route_pool_funcs = {IP4: v4_route_pool_from_ips, IP6: v6_route_pool_from_ips}

    # Return route pool based on version
    rp = {}
    for af in (IP4, IP6):
        # Get list of routes based on IPR inputs.
        routes = get_route_pool_funcs[af](ipr_list[af], nic)

        # If IPs were specified then use them.
        # Otherwise use the already existing route pool.
        if routes:
            rp[af] = RoutePool(routes)
        else:
            rp[af] = nic.rp[af]

    return rp


def sort_ips_by_nic(ip_list, nic_list):
    """Return a dict mapping each NIC id to the subset of ip_list addresses it owns."""
    by_nic = {}
    for nic in nic_list:
        by_nic[nic.id] = set()

    for ip in ip_list:
        ipr = IPR(ip)
        for nic in nic_list:
            # Search via route iteration (nic_ips, ext_ips, route-level link_locals).
            for route in nic.rp[ipr.af]:
                hey_stack = (
                    route.nic_ips,
                    route.ext_ips,
                    route.link_locals,
                )

                for hey in hey_stack:
                    if ipr in hey:
                        by_nic[nic.id].add(str(ipr))

            # Also check RoutePool-level link_locals directly.
            # These are populated during load_interface for every link-local on
            # the NIC, even when there are no global IPv6 routes to attach them
            # to (so the route iteration above yields nothing).
            for ll in nic.rp[ipr.af].link_locals:
                if ipr == ll:
                    by_nic[nic.id].add(str(ipr))

    # Convert back to list.
    by_nic_list = {}
    for nic_id in by_nic:
        by_nic_list[nic_id] = list(by_nic[nic_id])

    return by_nic_list


"""
async def workspace():
    ip_list = ["10.0.1.19", "fe80::20c:29ff:fe7d:6654", "fe80:0000:0000:0000:020c:29ff:fe7d:6653"]
    nic = await Interface()
    
    by_nic = sort_ips_by_nic(ip_list, [nic])
    print(by_nic)

    return
    rp = route_pool_from_ips(ip_list, nic)
    print(rp[IP4].to_dict())
    print(rp[IP6].to_dict())
    print("here")
    pass

if __name__ == "__main__":
    async_run(workspace())
"""



"""Simple echo server used for connectivity testing."""
import asyncio
from ...net.daemon import Daemon
from ...utility.utils import async_wrap_errors, get_running_loop
from ...utility.fstr import fstr
from ...net.asyncio.async_run import async_run


class EchoServer(Daemon):
    """Daemon that echoes every received message back to its sender."""
    def __init__(self):
        super().__init__()

    async def msg_cb(self, msg, client_tup, pipe):
        await async_wrap_errors(pipe.send(msg, client_tup))


if __name__ == "__main__":  # pragma: no cover

    class EchoProtocol(asyncio.Protocol):
        """asyncio.Protocol that logs connections and echoes received data back to the peer."""
        def connection_made(self, transport):
            """Store the transport and log the new incoming connection address."""
            self.transport = transport
            addr = transport.get_extra_info("peername")

        def data_received(self, data):
            """Echo the received data back to the sender and log the message."""
            message = data.decode()
            addr = self.transport.get_extra_info("peername")
            # Echo back
            self.transport.write(data)

        def connection_lost(self, exc):
            """Log the address of a peer whose connection has been closed."""
            addr = self.transport.get_extra_info("peername")

    async def echo_main():
        from aionetiface.src.aionetiface.net.net_utils import IP4, TCP
        from aionetiface.nic.interface import Interface

        loop = get_running_loop()
        server = await loop.create_server(lambda: EchoProtocol(), "127.0.0.1", 3000)

        async with server:
            await server.serve_forever()

        nic = await Interface()
        echo_route = await nic.route(IP4).bind(ips="localhost", port=3000)
        # print(echo_route)
        # print(echo_route._bind_tups)

        # Daemon instance.
        echod = EchoServer()
        await echod.add_listener(TCP, echo_route)

        while True:
            await asyncio.sleep(1)

    async_run(echo_main())



"""Constants and data structures for NAT detection."""
__all__ = [
    "STUN_PORT",
    "MAX_MAP_NO",
    "USE_MAP_NO",
    "OPEN_INTERNET",
    "SYMMETRIC_UDP_FIREWALL",
    "FULL_CONE",
    "RESTRICT_NAT",
    "RESTRICT_PORT_NAT",
    "SYMMETRIC_NAT",
    "BLOCKED_NAT",
    "NA_DELTA",
    "EQUAL_DELTA",
    "PRESERV_DELTA",
    "INDEPENDENT_DELTA",
    "DEPENDENT_DELTA",
    "RANDOM_DELTA",
    "EASY_NATS",
    "DELTA_N",
    "PREDICTABLE_NATS",
    "FUSSY_NATS",
    "BLOCKING_NATS",
]

STUN_PORT = 3478
MAX_MAP_NO = 100
USE_MAP_NO = 2

# NAT TYPES ---------------------------------------------

# No NAT at all.
OPEN_INTERNET = 1

# There is no NAT but there is some kind of firewall.
SYMMETRIC_UDP_FIREWALL = 2

# Mappings are made for local endpoints.
# Then any destination can use the mapping to reach the local enpoint.
# Note: May be incorrectly detected if using TCP.
FULL_CONE = 3

# NAT reuses mapping if same src ip and port is used.
# Destination must be white listed. It can use any port to send replies on.
# Endpoint-independent
# Note: May be incorrectly detected if using TCP.
RESTRICT_NAT = 4

# Mappings reused based on src ip and port.
# Destination must be white listed and use the port requested by recipient.
# Endpoint-independent (with some limitations.)
# Note: May be incorrectly detected if using TCP.
RESTRICT_PORT_NAT = 5

# Different mapping based on outgoing hosts.
# Even if same source IP and port reused.
# AKA: End-point dependent mapping.
SYMMETRIC_NAT = 6

# No response at all.
BLOCKED_NAT = 7
# ---------------------------------------------------------

# DELTA types: ------------------
# Mappings are easy to reuse and reach.
NA_DELTA = 1  # Not applicable.
EQUAL_DELTA = 2
PRESERV_DELTA = 3  # Or not applicable like in open internet.
INDEPENDENT_DELTA = 4
DEPENDENT_DELTA = 5
RANDOM_DELTA = 6
# -------------------------------

EASY_NATS = [OPEN_INTERNET, FULL_CONE]
DELTA_N = [
    # Remote = Local.
    EQUAL_DELTA,
    # Remote x - y = local x - y
    PRESERV_DELTA,
    # Remote x; y = remote x + delta, local = anything
    INDEPENDENT_DELTA,
    # Remote x; y = remote x + delta only when local x + delta
    DEPENDENT_DELTA,
]

# The NATs here have various properties that allow their
# mappings to be predictable under certain conditions.
PREDICTABLE_NATS = [
    # Once open - anyone can use the mapping.
    FULL_CONE,
    # Same local IP + port = same mapping.
    RESTRICT_NAT,
    # Same as above but reply port needs to match original dest.
    RESTRICT_PORT_NAT,
    # Open doesn't really apply so yes.
    OPEN_INTERNET,
]

# NAT types that require specific reply ports.
FUSSY_NATS = [
    RESTRICT_PORT_NAT,
]

# Peer will be unreachable.
BLOCKING_NATS = [BLOCKED_NAT]
